"""Implements the `json` domain and `json:schema` directive."""

import dataclasses
import json
import os
import re
from typing import (
    List,
    Any,
    Optional,
    Dict,
    cast,
    NamedTuple,
    Tuple,
    Iterator,
    Set,
    Union,
)

import docutils
import docutils.parsers.rst.directives

try:
    import jsonschema.validators

    _jsonschema_validation_supported = True
except ImportError:
    _jsonschema_validation_supported = False

import sphinx
import sphinx.addnodes
import sphinx.application
import sphinx.directives
import sphinx.domains
import sphinx.environment
import sphinx.project
import sphinx.roles
import sphinx.util
import sphinx.util.docutils
import sphinx.util.logging
import sphinx.util.matching
import yaml  # pylint: disable=import-error

from .. import object_description_options
from . import json_pprint
from ... import sphinx_utils
from ... import default_literal_role
from .. import apigen_utils

logger = sphinx.util.logging.getLogger(__name__)


class JsonSchemaMapEntry(NamedTuple):
    path: str
    pointer: str
    schema: Any
    id: str
    top_level_schema: Any


JsonSchema = Any
"""JSON schema represented as a JSON object (i.e. dict)."""

YamlSourceInfoMap = Dict[int, Tuple[str, int]]
"""For each literal string parsed from a schema YAML file, maps `id(s)` to the
`(source_file, lineno)`.

When embedding strings from the YAML files as reStructuredText content in the
documentation, this is used to include proper source information to Sphinx,
which in turn can include it in error messages.
"""


@dataclasses.dataclass
class LoadedSchemaData:
    id_map: Dict[str, JsonSchemaMapEntry] = dataclasses.field(default_factory=dict)
    """Maps JSON schema ids to the loaded schema entry.

    This contains an entry for both top-level and sub-schemas, including schemas
    for individual properties.

    The same schema entry may be contained under multiple keys.  In addition to
    the canonical id (defined by the $id member), if any, the schema is also
    present under a JSON pointer key `<parent>#/json/pointer/path`, if applicable.

    Additionally, as schemas are rendered to a document tree, additional aliases
    are added for member property ids `<parent>.<member>` and for string constant
    oneOfs `<parent>.<value>`.
    """

    source_info_map: YamlSourceInfoMap = dataclasses.field(default_factory=dict)
    """Combined source map for all schemas loaded in `id_map`."""

    identity_map: Dict[int, JsonSchemaMapEntry] = dataclasses.field(
        default_factory=dict
    )
    """Maps `id(entry.schema)` -> `entry` for every entry in `id_map`.

    This allows the entry to be determined from arbitrary nested schemas.
    """

    supertype_map: Dict[str, List[str]] = dataclasses.field(default_factory=dict)
    """Maps the canonical schema id to a list of types it "extends" (via allOf)."""

    subtype_map: Dict[str, List[str]] = dataclasses.field(default_factory=dict)
    """Maps the canonical schema id to a list of types that "extend" it).

    This is the inverse of `supertype_map`.
    """

    def register_subtype_relationships(self, schema: JsonSchema) -> None:
        """Registers subtype relationships involving `schema`.

        These are stored in the `subtype_map` and `schema_supertype_map` maps.

        This relationship information is used to render the "Extends" and "Subtypes"
        sections for a JSON schema description.

        :param schema: Schema to process.

        """
        schema_id = schema["$id"]
        # Don't show types with ids containing "#"
        if "#" in schema_id:
            return
        allof = schema.get("allOf")
        if not allof:
            return
        for supertype in allof:
            supertype_id = supertype.get("$ref")
            if supertype_id is None:
                continue
            supertype_entry = self.id_map[supertype_id]
            if "$id" not in supertype_entry.schema:
                continue
            # Don't show types with ids containing "#"
            if "#" in supertype_entry.id:
                continue
            supertype_id = supertype_entry.id
            self.subtype_map.setdefault(supertype_id, []).append(schema_id)
            self.supertype_map.setdefault(schema_id, []).append(supertype_id)

    def load_json_schema(
        self, path: str, full_path: str, seen_ids: Dict[str, str], validate: bool
    ) -> None:
        """Loads a single JSON schema into the global registry.

        All relative schema ids are resolved to absolute ids.

        The top-level schema as well as any sub-schemas are included in the registry.

        :param path: Path relative to documentation source dir.
        :param full_path: Full path to schema file.
        :param seen_ids: Dictionary of schema ids to populate.
        """
        with open(full_path, "r", encoding="utf-8") as f:
            top_level_schema, source_info_map = yaml_load(f, source_path=full_path)
        self.source_info_map.update(source_info_map)
        if validate:
            jsonschema.validators.validator_for(top_level_schema).check_schema(
                top_level_schema
            )
        top_level_id = top_level_schema.get("$id")
        _fix_jsonschema_ids(
            top_level_schema, top_level_id, path, seen_ids, self.source_info_map
        )

        # Register an individual schema/sub-schema.
        def _process_schema(schema: JsonSchema, pointer: str):
            pointer_based_id = top_level_id
            if pointer:
                pointer_based_id += "#" + pointer
            canonical_id = pointer_based_id
            if "$id" in schema:
                canonical_id = schema["$id"] = _normalize_jsonschema_id(
                    schema["$id"], top_level_id
                )
            schema_entry = JsonSchemaMapEntry(
                path=path,
                pointer=pointer,
                schema=schema,
                id=canonical_id,
                top_level_schema=top_level_schema,
            )
            self.identity_map[id(schema)] = schema_entry
            self.id_map[canonical_id] = schema_entry
            if canonical_id != pointer_based_id:
                self.id_map[pointer_based_id] = schema_entry

        for sub_schema, pointer in _traverse_sub_schemas(top_level_schema):
            _process_schema(sub_schema, pointer)


def yaml_load(  # pylint: disable=invalid-name
    stream,
    source_path: str,
) -> Tuple[Any, YamlSourceInfoMap]:
    """Loads a yaml file, preserving object key order and source line information.

    :param stream: File-like stream to read from.
    :param source_path: Path to source file for inclusion in source info map.
    :param Loader: YAML loader class.

    :returns: Tuple of loaded YAML value and source info map.
    """
    source_info_map: YamlSourceInfoMap = {}

    class Loader(yaml.SafeLoader):
        def compose_scalar_node(self, anchor):
            line = self.line
            node = super().compose_scalar_node(anchor)
            if isinstance(node.value, str):
                source_info_map[id(node.value)] = (source_path, line)
            return node

    result = yaml.load(stream, Loader)
    return result, source_info_map


def _globs_to_re(globs: List[str]) -> re.Pattern:
    return re.compile(
        "|".join(
            "(?:" + sphinx.util.matching._translate_pattern(x) + ")" for x in globs
        )
    )


def _get_json_schema_files(app: sphinx.application.Sphinx):
    """Finds all schema files.

    :param app: Sphinx application object.

    :Yields: Each schema path relative to ``app.srcdir``.
    """
    include_globs = app.config.json_schemas
    if not include_globs:
        return

    include_re = _globs_to_re(include_globs)

    exclude_paths = app.config.exclude_patterns + app.config.templates_path
    exclude_re = _globs_to_re(exclude_paths + sphinx.project.EXCLUDE_PATHS)

    matching_files = sphinx.util.get_matching_files(
        app.srcdir, (lambda s: exclude_re.fullmatch(s) is not None,)
    )
    for name in matching_files:
        if not include_re.fullmatch(name):
            continue
        yield name


def _populate_json_schema_id_map(app: sphinx.application.Sphinx):
    """Finds all schema files and loads them into `_json_schema_id_map`."""
    schema_data = LoadedSchemaData()
    setattr(app.env, "json_schema_data", schema_data)
    seen_ids: Dict[str, str] = {}
    all_paths = list(_get_json_schema_files(app))
    validate = app.config.json_schema_validate
    if all_paths and validate and not _jsonschema_validation_supported:
        logger.error("jsonschema package required for json_schema_validate option")
        validate = False
    for i, path in enumerate(all_paths):
        logger.info("[%d/%d] Loading JSON schema: %s", i + 1, len(all_paths), path)
        schema_data.load_json_schema(
            path=path,
            full_path=os.path.join(app.srcdir, path),
            seen_ids=seen_ids,
            validate=validate,
        )
    for key, path in seen_ids.items():
        if key in schema_data.id_map:
            continue
        logger.error("Undefined jsonschema $ref: %s: %s", path, key)

    for schema_id, schema_entry in schema_data.id_map.items():
        if schema_entry.id != schema_id:
            # Don't process under an alias, in order to avoid processing the same type
            # more than once.
            continue
        schema = schema_entry.schema
        if "$id" not in schema:
            continue
        schema_data.register_subtype_relationships(schema)


def _is_object_with_properties(schema_node: JsonSchema) -> bool:
    """Returns `True` if `schema_node` is an object schema with properties.

    An `allOf` schema is also assumed to be an object schema.

    :param schema_node: JSON schema to check.
    """
    if not isinstance(schema_node, dict):
        return False
    return ("type" not in schema_node and "allOf" in schema_node) or (
        schema_node.get("type") == "object" and schema_node.get("properties", {})
    )


def _is_object_array_with_properties(schema_node: JsonSchema):
    return schema_node.get("type") == "array" and _is_object_with_properties(
        schema_node.get("items")
    )


def _is_object_or_object_array_with_properties(schema_node: JsonSchema):
    return _is_object_with_properties(schema_node) or _is_object_array_with_properties(
        schema_node
    )


def _normalize_jsonschema_id(id_str: str, top_level_id: Optional[str]) -> str:
    if id_str.startswith("#") and top_level_id is not None:
        return top_level_id + id_str
    return id_str


def _traverse_sub_schemas(
    schema: JsonSchema, pointer: str = ""
) -> Iterator[Tuple[JsonSchema, str]]:
    """Invokes `callback` on `schema` and any sub-schemas.

    :param schema: JSON schema.
    :param pointer: Initial JSON pointer.
    """
    yield (schema, pointer)

    def _handle_prop(prop):
        obj = schema.get(prop)
        if obj is None:
            return
        if isinstance(obj, list):
            items = enumerate(obj)
        elif isinstance(obj, dict):
            items = obj.items()
        else:
            raise TypeError
        for key, sub_schema in items:
            yield from _traverse_sub_schemas(
                sub_schema, "%s/%s/%s" % (pointer, prop, key)
            )

    for prop in ("definitions", "properties", "allOf", "oneOf"):
        yield from _handle_prop(prop)
    items_data = schema.get("items")
    if items_data is not None:
        if isinstance(items_data, list):
            yield from _handle_prop("items")
        else:
            yield from _traverse_sub_schemas(items_data, "%s/items" % (pointer,))


def _fix_jsonschema_ids(
    schema: JsonSchema,
    top_level_id: Optional[str],
    path: str,
    seen_ids: Dict[str, str],
    source_info: YamlSourceInfoMap,
):
    """Makes relative `$id` and `$ref` members absolute.

    Also collects all ids that are seen in order to warn about unresolved
    references.

    :param schema: JSON schema, modified in place.
    :param top_level_id: Top-level id to use for resolving relative ids.
    :param path: Path to the file containing `schema`.
    :param seen_ids: Dictionary mapping seen ids to paths.
    :param source_info: Source info map to update.
    """

    def _fix_ids(sub_schema):
        for prop in ("$id", "$ref"):
            if prop in sub_schema:
                value = sub_schema[prop]
                normalized = _normalize_jsonschema_id(value, top_level_id)
                info = source_info.get(id(value))
                if info is not None:
                    source_info[id(normalized)] = info
                seen_ids.setdefault(normalized, path)
                sub_schema[prop] = normalized

    for sub_schema, pointer in _traverse_sub_schemas(schema):
        _fix_ids(sub_schema)


def _get_json_schema_node_id(fully_qualified_name: str) -> str:
    """Returns the reference id (i.e. HTML fragment id) for a schema."""
    return "json-%s" % (fully_qualified_name,)


def _schema_to_xref(schema_id: str) -> sphinx.addnodes.pending_xref:
    """Creates a pending_xref node that references a json schema.

    The result is similar to that of the json:schema role but just a bare
    pending_xref is constructed, without a code element.

    :param schema_id: Schema id.

    :returns: The pending_xref node.
    """
    return sphinx.addnodes.pending_xref(
        "",
        docutils.nodes.Text(schema_id),
        refdomain="json",
        reftype="schema",
        reftarget=schema_id,
    )


def _json_literal(self, j: Any) -> docutils.nodes.Node:
    return docutils.nodes.literal(
        text=json.dumps(j),
        classes=["code", "json"],
        language="json",
    )


class JsonSchemaDirective(sphinx.directives.ObjectDescription):
    """json:schema directive."""

    required_arguments = 1
    has_content = False

    objtype = "schema"

    option_spec = {
        "fully_qualified_name": str,
        "title": str,
        "toc_title": str,
        "exclude_from_toc": docutils.parsers.rst.directives.flag,
        "nested": docutils.parsers.rst.directives.flag,
        "noindex": docutils.parsers.rst.directives.flag,
    }

    _rendered_title: Optional[docutils.nodes.inline]

    def _get_schema_entry(self) -> JsonSchemaMapEntry:
        schema_data: LoadedSchemaData = getattr(self.env, "json_schema_data")
        return schema_data.id_map[self.arguments[0]]

    def _get_schema(self):
        return self._get_schema_entry().schema

    def _inline_text(self, text: str) -> List[docutils.nodes.Node]:
        nodes, messages = self.state.inline_text(text, self.lineno)
        return nodes + cast(List[docutils.nodes.Node], messages)

    def _json_sig_type(self, name: str):
        return sphinx.addnodes.desc_type(name, name)

    def _get_type_description_line_ref(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        return [
            sphinx.addnodes.desc_type("", "", _schema_to_xref(schema_node.get("$ref")))
        ]

    def _get_type_description_line_oneof(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        result: List[docutils.nodes.Node] = []
        for x in schema_node.get("oneOf"):
            if result:
                result.append(sphinx.addnodes.desc_sig_punctuation("", " | "))
            part = self._get_type_description_line(x)
            if part is None:
                return None
            result += part
        return result

    def _get_type_description_line_const(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        return [_json_literal(self.state, schema_node["const"])]

    def _get_type_description_line_enum(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        result: List[docutils.nodes.Node] = []
        for x in schema_node.get("enum"):
            if result:
                result.append(sphinx.addnodes.desc_sig_punctuation("", " | "))
            result.append(_json_literal(self.state, x))
        return result

    def _get_type_description_line_number(
        self, schema_node: JsonSchema, t: str
    ) -> Optional[List[docutils.nodes.Node]]:
        explicit_lower = True
        if schema_node.get("minimum") is not None:
            lower_punct = "["
            lower_node = _json_literal(self.state, schema_node.get("minimum"))
        elif schema_node.get("exclusiveMinimum") is not None:
            lower_punct = "("
            lower_node = _json_literal(self.state, schema_node.get("exclusiveMinimum"))
        else:
            lower_punct = "("
            lower_node = sphinx.addnodes.desc_sig_operator("", "-∞")
            explicit_lower = False

        explicit_upper = True
        if schema_node.get("maximum") is not None:
            upper_punct = "]"
            upper_node = _json_literal(self.state, schema_node.get("maximum"))
        elif schema_node.get("exclusiveMaximum") is not None:
            upper_punct = ")"
            upper_node = _json_literal(self.state, schema_node.get("exclusiveMaximum"))
        else:
            upper_punct = ")"
            upper_node = sphinx.addnodes.desc_sig_operator("", "+∞")
            explicit_upper = False

        result = [self._json_sig_type(t)]
        if explicit_lower or explicit_upper:
            result.append(
                docutils.nodes.subscript(
                    "",
                    "",
                    sphinx.addnodes.desc_sig_punctuation("", lower_punct),
                    lower_node,
                    sphinx.addnodes.desc_sig_punctuation("", ", "),
                    upper_node,
                    sphinx.addnodes.desc_sig_punctuation("", upper_punct),
                )
            )
        return result

    def _get_type_description_line_boolean(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        return [self._json_sig_type("boolean")]

    def _get_type_description_line_string(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        result = [self._json_sig_type("string")]
        subscript_parts: List[docutils.nodes.Node] = []
        if (
            schema_node.get("minLength") is not None
            or schema_node.get("maxLength") is not None
        ):
            subscript_parts.append(sphinx.addnodes.desc_sig_punctuation("", "["))
            min_length = schema_node.get("minLength")
            if min_length:
                subscript_parts.append(_json_literal(self.state, min_length))
            subscript_parts.append(sphinx.addnodes.desc_sig_punctuation("", ".."))
            max_length = schema_node.get("maxLength")
            if max_length:
                subscript_parts.append(_json_literal(self.state, max_length))
            subscript_parts.append(sphinx.addnodes.desc_sig_punctuation("", "]"))
            result.append(docutils.nodes.subscript("", "", *subscript_parts))
        return result

    def _get_type_description_line_array(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        items = schema_node.get("items")
        prefix = [self._json_sig_type("array")]
        if "minItems" in schema_node or "maxItems" in schema_node:
            prefix.append(sphinx.addnodes.desc_sig_punctuation("", "["))
            if schema_node.get("minItems") == schema_node.get("maxItems"):
                prefix.append(_json_literal(self.state, schema_node["minItems"]))
            else:
                if schema_node["minItems"]:
                    prefix.append(_json_literal(self.state, schema_node["minItems"]))
                prefix.append(sphinx.addnodes.desc_sig_punctuation("", ".."))
                if schema_node["maxItems"]:
                    prefix.append(_json_literal(self.state, schema_node["maxItems"]))
            prefix.append(sphinx.addnodes.desc_sig_punctuation("", "]"))

        if "items" in schema_node and isinstance(items, dict):
            items_desc = self._get_type_description_line(items)
            if items_desc is None:
                return None
            return prefix + [docutils.nodes.emphasis("", " of ")] + items_desc
        if "items" in schema_node and isinstance(items, list):
            result: List[docutils.nodes.Node] = []
            result.append(sphinx.addnodes.desc_sig_punctuation("", "["))
            for i, item in enumerate(items):
                if i != 0:
                    result.append(sphinx.addnodes.desc_sig_punctuation("", ","))
                    result.append(docutils.nodes.Text(" "))
                item_desc = self._get_type_description_line(item)
                if item_desc is None:
                    return prefix
                result.extend(item_desc)
            result.append(sphinx.addnodes.desc_sig_punctuation("", "]"))
            return result
        return prefix

    def _get_type_description_line_null(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        return [_json_literal(self.state, None)]

    def _get_type_description_line_object(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        return [self._json_sig_type("object")]

    def _get_type_description_line(
        self, schema_node: JsonSchema
    ) -> Optional[List[docutils.nodes.Node]]:
        """Renders a short type description for use in the signature."""
        if "$ref" in schema_node:
            return self._get_type_description_line_ref(schema_node)
        if "oneOf" in schema_node:
            return self._get_type_description_line_oneof(schema_node)
        if "const" in schema_node:
            return self._get_type_description_line_const(schema_node)
        if "enum" in schema_node:
            return self._get_type_description_line_enum(schema_node)
        t = schema_node.get("type")
        if "allOf" in schema_node and t is None:
            t = "object"
        if t in ("integer", "number"):
            return self._get_type_description_line_number(schema_node, t)
        if t == "boolean":
            return self._get_type_description_line_boolean(schema_node)
        if t == "string":
            return self._get_type_description_line_string(schema_node)
        if t == "array":
            return self._get_type_description_line_array(schema_node)
        if t == "null":
            return self._get_type_description_line_null(schema_node)
        if t == "object":
            return self._get_type_description_line_object(schema_node)
        return None

    def _collect_object_properties(
        self,
        schema_node: JsonSchema,
        properties: Dict[str, JsonSchema],
        required: Set[str],
    ) -> None:
        """Collects the object properties for a schema.

        :param schema_node: JSON schema to process.
        :param properties: Dict to be filled with members.
        :param required: Set to which required members are added.
        """
        schema_data: LoadedSchemaData = getattr(self.env, "json_schema_data")
        if "$ref" in schema_node:
            schema_node = schema_data.id_map[schema_node["$ref"]].schema
        if schema_node.get("type") == "object":
            properties.update(schema_node.get("properties", {}))
            required.update(schema_node.get("required", []))
        elif _is_object_array_with_properties(schema_node):
            self._collect_object_properties(schema_node["items"], properties, required)
        else:
            for sub_node in schema_node.get("allOf", []):
                self._collect_object_properties(sub_node, properties, required)

    def _add_sub_schema(
        self, entry: JsonSchemaMapEntry, **kwargs: Union[str, bool, None]
    ) -> List[docutils.nodes.Node]:
        """Renders a nested schema.

        :param entry: Entry for the nested schema.
        :param kwargs: Options for the json:schema directive.
        :returns: The rendered result as a list of docutils nodes.
        """
        return sphinx_utils.parse_rst(
            self.state,
            sphinx_utils.format_directive("json:schema", entry.id, options=kwargs),
            *self.get_source_info(),
        )

    def _make_field(
        self, label: str
    ) -> Tuple[docutils.nodes.field_list, docutils.nodes.field_body]:
        field_list = docutils.nodes.field_list()
        field = docutils.nodes.field()
        field_list += field
        field_name = docutils.nodes.field_name(label, label)
        field += field_name
        if not self._noindex and self._objdesc_options["include_fields_in_toc"]:
            field_name["ids"].append(
                docutils.nodes.make_id(label) + "-" + cast(str, self._node_id)
            )
            field_name["toc_title"] = label
        body = docutils.nodes.field_body(classes=["noindent"])
        field += body
        return field_list, body

    def _render_oneofs(self) -> List[docutils.nodes.Node]:
        """Renders a long-form list of oneof sub-schemas.

        :returns: The rendered result as a list of docutils nodes.
        """
        schema_data: LoadedSchemaData = getattr(self.env, "json_schema_data")
        field_list, body = self._make_field("One of")
        one_ofs = self._schema_entry.schema["oneOf"]
        # If all oneof options are constant strings, generate fully-qualified
        # ids.
        generate_oneof_ids = all(isinstance(x.get("const"), str) for x in one_ofs)
        for x in one_ofs:
            fully_qualified_name = None
            title = None
            if generate_oneof_ids:
                title = x["const"]
                fully_qualified_name = "%s.%s" % (self._fully_qualified_name, title)
            body.extend(
                self._add_sub_schema(
                    schema_data.identity_map[id(x)],
                    nested=True,
                    fully_qualified_name=fully_qualified_name,
                    exclude_from_toc=self._exclude_from_toc or not generate_oneof_ids,
                    noindex=self._noindex,
                    toc_title=title,
                )
            )
        return [field_list]

    def _get_source_info_from_schema_string(
        self, source_string: str
    ) -> Tuple[str, int]:
        """Obtains the source information for a schema string.

        If no source information is found, just returns the path of the current
        schema file without a valid line number.

        :param source_string: String literal obtained from a JSON schema.
        :returns: A tuple (path, line) specifying the source information.
        """

        schema_data: LoadedSchemaData = getattr(self.env, "json_schema_data")
        source_info = schema_data.source_info_map.get(id(source_string))
        if source_info is None:
            source_info = (self._schema_entry.path, -1)
        return source_info

    def _parse_rst_from_schema_string(
        self, source_string: str
    ) -> List[docutils.nodes.Node]:
        """Renders RST markup obtained from a schema.

        :param source_string: String literal obtained from a JSON schema.

        :returns: The rendered result as a list of docutils nodes.
        """
        rst_input = docutils.statemachine.StringList()
        sphinx_utils.append_multiline_string_to_stringlist(
            rst_input,
            ".. highlight-push::\n\n" + self.env.config.json_schema_rst_prolog + "\n\n",
            "<json_schema_rst_prolog>",
            -2,
        )
        sphinx_utils.append_multiline_string_to_stringlist(
            rst_input,
            source_string,
            *self._get_source_info_from_schema_string(source_string),
        )
        sphinx_utils.append_multiline_string_to_stringlist(
            rst_input,
            "\n\n"
            + self.env.config.json_schema_rst_epilog
            + "\n\n.. highlight-pop::\n\n",
            "<json_schema_rst_prolog>",
            -2,
        )
        with apigen_utils.save_rst_defaults(self.env):
            return sphinx_utils.parse_rst(
                self.state,
                rst_input,
                *self._get_source_info_from_schema_string(source_string),
            )

    def _render_body(self):
        """Renders the body of the schema."""
        schema_data: LoadedSchemaData = getattr(self.env, "json_schema_data")
        schema_node = self._schema_entry.schema
        result = []

        if self._rendered_title is not None:
            p = self._rendered_title
            p += docutils.nodes.Text("  ")
            result.append(p)
        if "description" in schema_node:
            result += self._parse_rst_from_schema_string(schema_node.get("description"))

        def _render_related_type_list(related: Optional[List[str]], label: str) -> None:
            if not related:
                return
            field_list, body = self._make_field(label)
            result.append(field_list)
            bullet_list = docutils.nodes.bullet_list()
            body += bullet_list
            for schema_id in related:
                related_schema = schema_data.id_map[schema_id].schema
                list_item = docutils.nodes.list_item()
                bullet_list += list_item
                p = docutils.nodes.paragraph()
                list_item += p
                p += self._inline_text(f":json:schema:`{schema_id}`")
                related_title = related_schema.get("title")
                if related_title:
                    p += self._inline_text(f" — {related_title}")

        _render_related_type_list(
            schema_data.supertype_map.get(self._schema_entry.id), "Extends"
        )
        subtypes = schema_data.subtype_map.get(self._schema_entry.id)
        if subtypes:
            subtypes = sorted(subtypes)
        _render_related_type_list(subtypes, "Subtypes")

        if "oneOf" in schema_node and any(
            ("title" in x or "description" in x) for x in schema_node["oneOf"]
        ):
            result.extend(self._render_oneofs())
        if _is_object_or_object_array_with_properties(schema_node):
            properties: Dict[str, JsonSchema] = {}
            required_properties: Set[str] = set()
            self._collect_object_properties(
                schema_node, properties, required_properties
            )
            for required in [True, False]:
                if not any(
                    (member_name in required_properties) == required
                    for member_name in properties
                ):
                    continue
                heading = "%s members" % ("Required" if required else "Optional")
                field_list, body = self._make_field(heading)
                result.append(field_list)
                for member_name, member_schema in properties.items():
                    if (member_name in required_properties) != required:
                        continue
                    body.extend(
                        self._add_sub_schema(
                            schema_data.identity_map[id(member_schema)],
                            title=member_name,
                            fully_qualified_name="%s.%s"
                            % (self._fully_qualified_name, member_name),
                            nested=True,
                            exclude_from_toc=self._exclude_from_toc,
                            noindex=self._noindex,
                        )
                    )

        def add_example(value, caption, admonition_class="example"):
            nonlocal result
            result += sphinx_utils.parse_rst(
                self.state,
                sphinx_utils.format_directive(
                    "admonition",
                    caption,
                    options={"class": admonition_class},
                    content=sphinx_utils.format_directive(
                        "code-block",
                        "json",
                        content=json_pprint.pformat(value, indent=2),
                    ),
                ),
                *self._get_source_info_from_schema_string(value),
            )

        if self._long_default_value:
            add_example(schema_node["default"], "Default", "note")

        if "examples" in schema_node:
            for example in schema_node["examples"]:
                add_example(example, "Example")
        return result

    def before_content(self):
        # Set ref_context information which will be used to resolve cross references
        # in the body.
        self.env.ref_context["json:schema"] = self._fully_qualified_name
        self.env.ref_context.setdefault("json:schemas", []).append(  # type: ignore
            self._fully_qualified_name
        )

    def after_content(self):
        # Clear the ref_context information set by `.before_context`.
        schemas = self.env.ref_context["json:schemas"]
        schemas.pop()
        self.env.ref_context["json:schema"] = schemas[-1] if schemas else None

    def transform_content(self, contentnode: sphinx.addnodes.desc_content) -> None:
        contentnode += self._render_body()

    def handle_signature(
        self, sig: str, signode: sphinx.addnodes.desc_signature
    ) -> Any:
        # Renders the header for the JSON schema.
        schema_entry = self._schema_entry
        if not self._nested:
            signode += sphinx.addnodes.desc_annotation("json ", "json ")
        type_line = self._type_line
        title = self._title
        if title is not None:
            signode += sphinx.addnodes.desc_name(title, title)
            if type_line:
                signode += sphinx.addnodes.desc_sig_punctuation("", " : ")
        signode += type_line
        if self._short_default_value:
            signode += sphinx.addnodes.desc_sig_punctuation("", " = ")
            signode += docutils.nodes.literal(
                text=self._short_default_value,
                classes=["code", "json"],
                language="json",
            )
        if not self._exclude_from_toc and self._toc_title:
            signode["toc_title"] = self._toc_title
        return schema_entry.id

    def add_target_and_index(
        self, name: Any, sig: str, signode: sphinx.addnodes.desc_signature
    ) -> None:
        del name
        if self._noindex:
            return
        signode["ids"].append(self._node_id)

    def run(self) -> List[docutils.nodes.Node]:

        schema_data: LoadedSchemaData = self.env.json_schema_data  # type: ignore
        schema_id = self.arguments[0]
        schema_entry = schema_data.id_map.get(schema_id)
        if schema_entry is None:
            logger.error(
                "Undefined JSON schema: %r", schema_id, location=self.get_source_info()
            )
            return []
        self._schema_entry = schema_entry

        self._fully_qualified_name = self.options.get("fully_qualified_name")
        self._noindex = "noindex" in self.options
        self._exclude_from_toc = "exclude_from_toc" in self.options
        self._nested = "nested" in self.options
        self._objtype = "schema" if not self._nested else "subschema"
        self._objdesc_options = (
            object_description_options.get_object_description_options(
                self.env, "json", self._objtype
            )
        )
        if self._fully_qualified_name:
            schema_data.id_map.setdefault(
                self._fully_qualified_name, self._schema_entry
            )
        else:
            self._fully_qualified_name = self._schema_entry.id

        if "#/" in self._fully_qualified_name:
            self._noindex = True

        if self._noindex:
            self._exclude_from_toc = True

        self._node_id = None
        if not self._noindex:
            self._node_id = _get_json_schema_node_id(self._fully_qualified_name)

        schema_node = self._schema_entry.schema
        type_line = self._get_type_description_line(schema_node)

        # Determine how to display the default value.  If it is short enough, it is
        # displayed inline after the type.  Otherwise, it is displayed in a separate
        # block after the description.
        self._long_default_value = None
        self._short_default_value = None
        has_default_value = "default" in schema_node
        if has_default_value:
            formatted_default = json_pprint.pformat(
                schema_node["default"], indent=2
            ).strip()
            if "\n" not in formatted_default and len(formatted_default) < 40:
                self._short_default_value = formatted_default
                has_default_value = False
            else:
                self._long_default_value = formatted_default
        self._type_line = type_line

        if "title" in self.options:
            self._title = self.options["title"]
        elif not self._nested:
            self._title = self._fully_qualified_name
        else:
            self._title = None

        self._toc_title = self.options.get("toc_title") or self._title

        schema = self._schema_entry.schema
        if "title" in schema:
            self._rendered_title = docutils.nodes.inline()
            self._rendered_title += self._parse_rst_from_schema_string(schema["title"])
        else:
            self._rendered_title = None

        if self._node_id:
            domain = cast(JsonSchemaDomain, self.env.get_domain("json"))
            generate_synopses = self._objdesc_options["generate_synopses"]
            synopsis: Optional[str]
            if generate_synopses is not None and self._rendered_title:
                synopsis = sphinx_utils.summarize_element_text(
                    self._rendered_title, generate_synopses
                )
            else:
                synopsis = None
            domain.schemas[self._fully_qualified_name] = DomainSchemaEntry(
                docname=self.env.docname,
                node_id=self._node_id,
                schema_id=self._schema_entry.id,
                synopsis=synopsis,
                objtype=self._objtype,
            )
        return super().run()


class JsonSchemaRole(sphinx.roles.XRefRole):
    """json:schema role."""

    def process_link(
        self,
        env: sphinx.environment.BuildEnvironment,
        refnode: docutils.nodes.Element,
        has_explicit_title: bool,
        title: str,
        target: str,
    ) -> Tuple[str, str]:
        # Based on PyXRefRole in sphinx/domains/python.py
        refnode["json:schema"] = env.ref_context.get("json:schema")
        if not has_explicit_title:
            title = title.lstrip(".")  # Only has a meaning for the target
            target = target.lstrip("~")  # Only has a meaning for the title
            # If the first character is a tilde, don't display the initial parts
            if title[0:1] == "~":
                title = title[1:]
                dot = title.rfind(".")
                if dot != -1:
                    title = title[dot + 1 :]
        # if the first character is a dot, search more specific namespaces first
        # else search builtins first
        if target[0:1] == ".":
            target = target[1:]
            refnode["refspecific"] = True
        return title, target


class DomainSchemaEntry(NamedTuple):
    """Entry representing a schema in the JSON domain."""

    docname: str
    """Document id where schema is defined."""

    node_id: str
    """Sphinx reference id"""

    schema_id: str
    """Canonical schema id"""

    synopsis: Optional[str]
    """Rendered text content of synopsis (title property)"""

    objtype: str
    """Object type, either 'schema' or 'subschema'"""


# Sphinx object priority values

OBJECT_PRIORITY_DEFAULT = 1
OBJECT_PRIORITY_IMPORTANT = 0
OBJECT_PRIORITY_UNIMPORTANT = 2
OBJECT_PRIORITY_EXCLUDE_FROM_SEARCH = -1


class JsonSchemaDomain(sphinx.domains.Domain):
    """Domain for JSON schemas."""

    name = "json"
    label = "JSON"

    roles = {
        "schema": JsonSchemaRole(warn_dangling=True),
    }

    object_types: Dict[str, sphinx.domains.ObjType] = {
        "schema": sphinx.domains.ObjType("type", "schema"),
        "subschema": sphinx.domains.ObjType("member", "schema"),
    }

    directives = {
        "schema": JsonSchemaDirective,
    }

    initial_data: Dict[str, DomainSchemaEntry] = {
        "schemas": cast(DomainSchemaEntry, {}),
    }

    @property
    def schemas(self) -> Dict[str, DomainSchemaEntry]:
        return self.data["schemas"]

    def get_objects(self) -> Iterator[Tuple[str, str, str, str, str, int]]:
        for refname, obj in self.schemas.items():
            yield (
                refname,
                refname,
                obj.objtype,
                obj.docname,
                obj.node_id,
                OBJECT_PRIORITY_DEFAULT
                if obj.objtype == "schema"
                else OBJECT_PRIORITY_UNIMPORTANT,
            )

    def merge_domaindata(
        self, docnames: List[str], otherdata: Dict
    ) -> None:  # pylint: disable=g-bare-generic
        self.schemas.update(otherdata["schemas"])

    def _find_schema(
        self, target: str, parent_schema: Optional[str], refspecific: bool
    ) -> Optional[Tuple[str, DomainSchemaEntry]]:
        """Looks up a cross-referenced schema.

        :param target: Target schema identifier.
        :param parent_schema: Fully-qualified parent id.
        :param refspecific: Whether to search for relative targets first.

        :returns:
          If not found, returns `None`.  Otherwise, returns the
          ``(fully_qualified_id, entry)``
        """
        search_prefixes = [""]
        if parent_schema:
            parent_parts = parent_schema.split(".")
            for i in range(1, len(parent_parts) + 1):
                search_prefixes.append(".".join(parent_parts[:i]) + ".")
        if refspecific:
            search_prefixes = search_prefixes[::-1]
        for search_prefix in search_prefixes:
            full_name = search_prefix + target
            domain_entry = self.schemas.get(full_name)
            if domain_entry is not None:
                return (full_name, domain_entry)
        # Match not found
        return None

    def resolve_xref(
        self,
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        typ: str,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> Optional[docutils.nodes.Element]:
        del typ
        match = self._find_schema(
            target=target,
            parent_schema=node.get("json:schema"),
            refspecific=node.hasattr("refspecific"),
        )
        if match is None:
            return None
        return self._make_refnode(
            builder=builder, fromdocname=fromdocname, contnode=contnode, match=match
        )

    def _make_refnode(
        self,
        builder: sphinx.builders.Builder,
        fromdocname: str,
        contnode: docutils.nodes.Element,
        match: Tuple[str, DomainSchemaEntry],
    ) -> docutils.nodes.Element:
        full_name, domain_entry = match
        options = object_description_options.get_object_description_options(
            self.env, "json", domain_entry.objtype
        )
        tooltip = object_description_options.format_object_description_tooltip(
            self.env, options, full_name, domain_entry.synopsis
        )
        return sphinx.util.nodes.make_refnode(
            builder=builder,
            fromdocname=fromdocname,
            todocname=domain_entry.docname,
            targetid=domain_entry.node_id,
            child=contnode,
            title=tooltip,
        )

    def resolve_any_xref(
        self,
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> List[Tuple[str, docutils.nodes.Element]]:
        match = self._find_schema(
            target=target, parent_schema=node.get("json:schema"), refspecific=True
        )
        results: List[Tuple[str, docutils.nodes.Element]] = []
        if match is not None:
            results.append(
                (
                    "json:schema",
                    self._make_refnode(
                        builder=builder,
                        fromdocname=fromdocname,
                        contnode=contnode,
                        match=match,
                    ),
                )
            )
        return results

    def get_object_synopses(self) -> Iterator[Tuple[Tuple[str, str], str]]:
        for obj in self.schemas.values():
            synopsis = obj.synopsis
            if synopsis:
                yield ((obj.docname, obj.node_id), synopsis)


def _builder_inited(app: sphinx.application.Sphinx):
    _populate_json_schema_id_map(app)


def setup(app: sphinx.application.Sphinx):
    app.add_domain(JsonSchemaDomain)
    app.add_config_value("json_schemas", types=(List[str],), default=[], rebuild="env")
    app.add_config_value(
        "json_schema_validate", types=(bool,), default=False, rebuild="env"
    )
    app.add_config_value(
        "json_schema_rst_prolog", types=(str,), default="", rebuild="env"
    )
    app.add_config_value(
        "json_schema_rst_epilog", types=(str,), default="", rebuild="env"
    )
    app.connect("builder-inited", _builder_inited)
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Sphinx extension that formats signatures using an external tool.

Currently only clang-format is supported.
"""

import collections
import hashlib
import io
import json
import re
import subprocess
from typing import Dict, List, Any, Union, Optional, cast, Tuple, Type

import docutils.nodes
import sphinx.addnodes
import sphinx.application
import sphinx.environment
import sphinx.transforms
import sphinx.util.logging

from . import object_description_options

logger = sphinx.util.logging.getLogger(__name__)

_SIGNATURE_FORMAT_ID = "signature_format_id"

_FORMATTED_SIGNATURES = "sphinx_immaterial_formatted_signatures"


def _get_collected_signatures(
    env: sphinx.environment.BuildEnvironment,
) -> Dict[Tuple[str, str], Dict[str, str]]:
    attr = "_sphinx_immaterial_collected_signatures_to_format"
    x = getattr(env, attr, None)
    if x is None:
        x = collections.defaultdict(dict)
        setattr(env, attr, x)
    return x


class SignatureText(docutils.nodes.Text):
    pass


def visit_signature_text(self, node: SignatureText) -> None:
    encoded = self.encode(node.astext()).replace(" ", "&nbsp;").replace("\n", "<br>")
    self.body.append(encoded)


def depart_signature_text(self, node: SignatureText) -> None:
    pass


class CollectSignaturesTransform(sphinx.transforms.SphinxTransform):

    # Late
    default_priority = 900

    def apply(self, **kwargs: Any) -> None:
        collected_signatures = _get_collected_signatures(self.env)
        for node in self.document.findall(sphinx.addnodes.desc_signature):
            parent = node.parent
            domain: str = parent.get("domain")
            objtype: str = parent.get("objtype")
            options = object_description_options.get_object_description_options(
                self.env, domain, objtype
            )
            if options.get("clang_format_style") is None:
                continue
            if "api-include-path" in node["classes"]:
                continue
            parts = []
            for child in node.children:
                parts.append(child.astext())
            signature = " ".join(parts)
            sig_id = hashlib.md5(
                (f"{domain}:{objtype}:" + signature).encode("utf-8")
            ).hexdigest()
            node[_SIGNATURE_FORMAT_ID] = sig_id
            collected_signatures[domain, objtype][sig_id] = signature


def _append_child_copy_source_info(
    target: docutils.nodes.Element,
    source_node: docutils.nodes.Node,
    node: docutils.nodes.Node,
):
    source = source_node.source
    line = source_node.line
    target.append(node)
    node.source = source
    node.line = line


def _extend_children_copy_source_info(
    target: docutils.nodes.Element,
    source_node: docutils.nodes.Node,
    children: List[docutils.nodes.Node],
):
    source = source_node.source
    line = source_node.line
    target.extend(children)
    for child in children:
        child.source = source
        child.line = line


def _format_signature(node: sphinx.addnodes.desc_signature, formatted: str) -> None:
    if "sig-wrap" in node["classes"]:
        node["classes"].remove("sig-wrap")
    children = list(node.children)
    del node[:]

    formatted_i = 0

    def format_next_text(text: str) -> str:
        nonlocal formatted_i
        if formatted[formatted_i : formatted_i + len(text)] == text:
            # Matches fully
            formatted_i += len(text)
            return text

        pattern = re.compile(
            r"\s*" + r"\s*".join(re.escape(x) for x in re.sub(r"\s", "", text))
        )
        m = pattern.match(formatted, formatted_i)
        if m is None:
            raise ValueError("Failed to match")
        formatted_i = m.end()
        return m.group(0)

    prev_text_node = None
    seen_name = False
    name_text_node_replacement = None

    def process_text(node: docutils.nodes.Text):
        nonlocal prev_text_node
        new_text = format_next_text(node.astext())
        new_node = SignatureText(new_text)
        if not seen_name:
            prev_text_node = new_node
        new_node.source = node.source
        new_node.line = node.line
        return [new_node]

    def process_desc_signature_line(node: sphinx.addnodes.desc_signature_line):
        # Exclude the `desc_signature_line` element and just include its children
        # directly.
        new_children = []
        for child in node.children:
            new_children.extend(process(child))
        return new_children

    def process_pending_xref(node: sphinx.addnodes.pending_xref):
        node = cast(sphinx.addnodes.pending_xref, node)
        old_children = list(node.children)
        del node[:]
        for child in old_children:
            _extend_children_copy_source_info(node, child, process(child))
        return [node]

    def convert_desc_parameter(
        node: sphinx.addnodes.desc_parameter,
    ) -> docutils.nodes.inline:
        new_node = docutils.nodes.inline("", "")
        new_node["classes"] = node["classes"]
        for child in node.children:
            _append_child_copy_source_info(new_node, child, child)
        new_node.source = node.source
        new_node.line = node.line
        return new_node

    def convert_desc_parameterlist(
        node: sphinx.addnodes.desc_parameterlist,
    ) -> docutils.nodes.inline:
        # First replace with other representation
        new_node = docutils.nodes.inline("", "")
        new_node += sphinx.addnodes.desc_sig_punctuation("(", "(")
        for i, child in enumerate(node.children):
            if i != 0:
                new_node += sphinx.addnodes.desc_sig_punctuation(",", ",")
            _append_child_copy_source_info(new_node, child, child)
        new_node += sphinx.addnodes.desc_sig_punctuation(")", ")")
        return new_node

    def process_text_element(node: docutils.nodes.TextElement):
        assert node.child_text_separator == ""
        old_children = list(node.children)
        del node[:]
        for child in old_children:
            _extend_children_copy_source_info(node, child, process(child))
        return [node]

    def process(node: docutils.nodes.Node):
        nonlocal seen_name, name_text_node_replacement
        if isinstance(node, docutils.nodes.Text):
            return process_text(node)
        if not seen_name:
            if isinstance(node, sphinx.addnodes.desc_addname) or (
                isinstance(node, sphinx.addnodes.desc_name)
                and "sig-name-nonprimary"
                not in cast(docutils.nodes.Element, node)["classes"]
            ):
                seen_name = True
                if prev_text_node is not None:
                    m = re.fullmatch(
                        r"(.*\n)(\s+)$", prev_text_node.astext(), re.DOTALL
                    )
                    if m is not None:
                        name_text_node_replacement = (
                            prev_text_node,
                            SignatureText(m.group(1)),
                        )

        if isinstance(node, sphinx.addnodes.desc_signature_line):
            return process_desc_signature_line(node)
        if isinstance(node, sphinx.addnodes.pending_xref):
            return process_pending_xref(node)
        if isinstance(node, sphinx.addnodes.desc_parameter):
            node = convert_desc_parameter(node)
        elif isinstance(node, sphinx.addnodes.desc_parameterlist):
            node = convert_desc_parameterlist(node)
        if isinstance(node, docutils.nodes.TextElement):
            return process_text_element(node)
        raise ValueError("unexpected child")

    node["is_multiline"] = False
    for child in children:
        _extend_children_copy_source_info(node, child, process(child))

    if name_text_node_replacement is not None:
        (  # pylint: disable=unpacking-non-sequence
            existing_node,
            new_node,
        ) = name_text_node_replacement
        source = existing_node.source
        line = existing_node.line
        existing_node.parent.replace(existing_node, new_node)
        new_node.source = source
        new_node.line = line


class FormatSignaturesTransform(sphinx.transforms.SphinxTransform):

    # Early
    default_priority = 0

    def apply(self, **kwargs: Any) -> None:
        formatted_signatures = getattr(self.env, _FORMATTED_SIGNATURES, None)
        if formatted_signatures is None:
            return
        for node in self.document.findall(sphinx.addnodes.desc_signature):
            signature_id = node.get(_SIGNATURE_FORMAT_ID)
            if signature_id is None:
                continue
            formatted_signature = formatted_signatures.get(signature_id)
            if formatted_signature is None:
                continue
            _format_signature(node, formatted_signature)


def merge_info(
    app: sphinx.application.Sphinx,
    env: sphinx.environment.BuildEnvironment,
    docnames: List[str],
    other: sphinx.environment.BuildEnvironment,
) -> None:
    _get_collected_signatures(env).update(_get_collected_signatures(other))


DOMAIN_CLANG_FORMAT_LANGUAGE = {
    "cpp": "Cpp",
    "c": "Cpp",
    "js": "JavaScript",
}

ClangFormatStyle = Union[str, Dict[str, Any]]


def env_updated(
    app: sphinx.application.Sphinx, env: sphinx.environment.BuildEnvironment
) -> None:
    all_signatures = _get_collected_signatures(env)
    formatted_signatures = {}

    signatures_for_style: Dict[str, Dict[str, str]] = collections.defaultdict(dict)

    for (domain, objtype), signatures in all_signatures.items():
        options = object_description_options.get_object_description_options(
            env, domain, objtype
        )

        style: ClangFormatStyle = options["clang_format_style"]
        if isinstance(style, str):
            style = {"BasedOnStyle": style}
        else:
            style = style.copy()

        style.setdefault("ColumnLimit", options["wrap_signatures_column_limit"])
        language = DOMAIN_CLANG_FORMAT_LANGUAGE.get(domain)
        if language is not None:
            style.setdefault("Language", language)
        style_key = json.dumps(style, sort_keys=True)
        signatures_for_style[style_key].update(signatures)

    for style_key, signatures in signatures_for_style.items():
        source = io.StringIO()

        for sig_id, signature in signatures.items():
            source.write(f"// {sig_id}\n")
            source.write(signature.strip().strip(";"))
            source.write(";\n")

        result = subprocess.run(
            [app.config.clang_format_command, f"-style={style_key}"],
            input=source.getvalue(),
            encoding="utf-8",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=False,
        )
        if result.returncode != 0:
            logger.error(
                f"{app.config.clang_format_command} exited with code %d: %s",
                result.returncode,
                result.stderr,
            )
            return
        result.check_returncode()
        stdout = result.stdout

        for m in re.finditer(
            "^// ([0-9a-f]+)\n((?:(?!\n//).)+)", stdout, re.MULTILINE | re.DOTALL
        ):
            formatted_signatures[m.group(1)] = m.group(2)

    setattr(env, _FORMATTED_SIGNATURES, formatted_signatures)


def setup(app: sphinx.application.Sphinx):
    app.setup_extension("sphinx_immaterial.apidoc.object_description_options")
    app.add_transform(CollectSignaturesTransform)
    app.add_post_transform(FormatSignaturesTransform)
    object_description_options.add_object_description_option(
        app,
        "clang_format_style",
        default=None,
        type_constraint=Optional[ClangFormatStyle],
    )

    app.connect("env-merge-info", merge_info)
    app.connect("env-updated", env_updated)
    app.add_node(
        cast(Type[docutils.nodes.Element], SignatureText),
        html=(visit_signature_text, depart_signature_text),
    )
    app.add_config_value(
        name="clang_format_command",
        default="clang-format",
        rebuild="env",
        types=(str,),
    )
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Defines the `cpp_strip_namespaces_from_signatures` config option."""

from typing import List

import docutils.nodes
import sphinx.addnodes
import sphinx.application


def _strip_namespaces_from_signature(
    node: docutils.nodes.Element, namespaces: List[str]
):
    # Collect nodes to remove first, then remove them in reverse order.
    removals = []
    for child in node.findall(condition=sphinx.addnodes.desc_sig_name):
        parent = child.parent
        if not isinstance(parent, sphinx.addnodes.pending_xref):
            continue
        if (
            parent["reftype"] != "identifier"
            or parent["refdomain"] != "cpp"
            or parent["reftarget"] not in namespaces
        ):
            continue
        grandparent = parent.parent
        if isinstance(grandparent, sphinx.addnodes.desc_addname):
            continue
        index = grandparent.index(parent)
        if index + 1 >= len(grandparent.children):
            continue
        sibling = grandparent[index + 1]
        if (
            not isinstance(sibling, sphinx.addnodes.desc_sig_punctuation)
            or sibling.astext() != "::"
        ):
            continue
        removals.append((grandparent, index))

    removals.reverse()

    for (parent, index) in removals:
        del parent[index : index + 2]


def _strip_namespaces_from_signatures(
    app: sphinx.application.Sphinx,
    domain: str,
    objtype: str,
    content: docutils.nodes.Element,
) -> None:
    """object-description-transform callback that strips namespaces."""
    if domain != "cpp":
        return

    namespaces = app.config.cpp_strip_namespaces_from_signatures
    if not namespaces:
        return

    signatures = content.parent[:-1]
    for signode in signatures:
        assert isinstance(signode, sphinx.addnodes.desc_signature)
        _strip_namespaces_from_signature(signode, namespaces)


def setup(app: sphinx.application.Sphinx):
    app.add_config_value(
        "cpp_strip_namespaces_from_signatures",
        default=[],
        rebuild="env",
        types=(List[str],),
    )
    app.connect("object-description-transform", _strip_namespaces_from_signatures)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Utilities for use with Sphinx."""

import contextlib
import io
from typing import Optional, Union, List, Tuple, Mapping, Sequence

import docutils.nodes
import docutils.parsers.rst.roles
import docutils.parsers.rst.states
import docutils.statemachine
import sphinx.addnodes
import sphinx.util.docutils
from typing_extensions import Literal


def to_statemachine_stringlist(
    content: str, source_path: str, source_line: int = 0
) -> docutils.statemachine.StringList:
    """Converts to a docutils StringList with associated source info.

    All lines of `content` are assigned the same source info.

    :param content: Source text.
    :param source_path: Path to the source file, for error messages.
    :param source_line: Line number in source file, for error messages.

    :returns: The string list, which may be passed to `nested_parse`.
    """
    list_lines = docutils.statemachine.string2lines(content)
    items = [(source_path, source_line)] * len(list_lines)
    return docutils.statemachine.StringList(list_lines, items=items)


def format_directive(
    name: str,
    *args: str,
    signatures: Optional[Sequence[str]] = None,
    content: Optional[str] = None,
    options: Optional[Mapping[str, Union[None, str, bool]]] = None,
) -> str:
    """Formats a RST directive into RST syntax.

    :param name: Directive name, e.g. "json:schema".
    :param args: List of directive arguments.
    :param signatures: List of signatures, alternative to ``args``.
    :param content: Directive body content.
    :param options: Directive options.

    :returns: The formatted directive.
    """
    out = io.StringIO()
    out.write("\n\n")
    assert not args or not signatures
    if not signatures:
        signatures = ["".join(f" {arg}" for arg in args)]
    prefix = f".. {name}:: "
    for i, signature in enumerate(signatures):
        if i == 1:
            prefix = " " * len(prefix)
        out.write(prefix + signature + "\n")
    if options:
        for key, value in options.items():
            if value is False or value is None:  # pylint: disable=g-bool-id-comparison
                continue
            if value is True:  # pylint: disable=g-bool-id-comparison
                value = ""
            out.write(f"   :{key}: {value}\n")
    if content:
        out.write("\n")
        for line in content.splitlines():
            out.write(f"   {line}\n")
    out.write("\n")
    return out.getvalue()


def append_directive_to_stringlist(
    out: docutils.statemachine.StringList,
    name: str,
    *args: str,
    source_path: str,
    source_line: int,
    indent: int = 0,
    signatures: Optional[Sequence[str]] = None,
    content: Union[str, docutils.statemachine.StringList, None] = None,
    options: Optional[Mapping[str, Union[None, str, bool]]] = None,
) -> None:
    """Formats a RST directive into RST syntax and appends to a StringList.

    :param out: StringList to which directive should be appended.
    :param name: Directive name, e.g. "json:schema".
    :param args: List of directive arguments.
    :param source_path: Source path to associate with the first line of the directive
        (and content if specified as a string).
    :param source_line: Source line to associate with the first line of the directive
        (and content if specified as a string).
    :param indent: Base amount to indent every line.
    :param content: Directive body content.
    :param options: Directive options.
    """
    base_indent_str = " " * indent
    # Blank line at beginning to start new directive.
    out.append(base_indent_str, source_path, source_line)
    prefix = f"{base_indent_str}.. {name}:: "
    assert not args or not signatures
    if not signatures:
        signatures = ["".join(f" {arg}" for arg in args)]
    for i, signature in enumerate(signatures):
        if i == 1:
            prefix = " " * len(prefix)
        out.append(prefix + signature, source_path, source_line)
    content_indent_str = base_indent_str + "   "
    if options:
        for key, value in options.items():
            if value is False or value is None:  # pylint: disable=g-bool-id-comparison
                continue
            if value is True:  # pylint: disable=g-bool-id-comparison
                value = ""
            out.append(f"{content_indent_str}:{key}: {value}", source_path, source_line)
    if content is not None:
        out.append(base_indent_str, source_path, source_line)
        if isinstance(content, str):
            append_multiline_string_to_stringlist(
                out, content, source_path, source_line, prefix=content_indent_str
            )
        else:
            for source, offset, line in content.xitems():
                out.append(content_indent_str + line, source, offset)
    # Blank line at end to denote end of directive.
    out.append(base_indent_str, source_path, source_line)


def append_multiline_string_to_stringlist(
    out: docutils.statemachine.StringList,
    text: str,
    source_path: str,
    source_line: int,
    prefix: str = "",
) -> None:
    for i, line in enumerate(text.splitlines()):
        out.append(prefix + line, source_path, source_line + i)


def parse_rst(
    state: docutils.parsers.rst.states.RSTState,
    text: Union[str, docutils.statemachine.StringList],
    source_path: str = "",
    source_line: int = 0,
) -> List[docutils.nodes.Node]:
    content = (
        to_statemachine_stringlist(text, source_path, source_line)
        if isinstance(text, str)
        else text
    )
    with sphinx.util.docutils.switch_source_input(state, content):
        node = docutils.nodes.container()
        # necessary so that the child nodes get the right source/line set
        node.document = state.document
        state.nested_parse(content, 0, node)
    return node.children


def summarize_element_text(
    node: docutils.nodes.Element,
    mode: Literal["first_paragraph", "first_sentence"] = "first_paragraph",
) -> str:
    """Extracts a short text synopsis, e.g. for use as a tooltip."""

    # Recurisvely extract first paragraph
    while True:
        for p in node.findall(condition=docutils.nodes.paragraph):
            if p is node:
                continue
            node = p
            break
        else:
            break

    text = node.astext()
    if mode == "first_sentence":
        sentence_end = text.find(". ")
        if sentence_end != -1:
            text = text[: sentence_end + 1]
    text = text.replace("\n", " ")
    return text.strip()


def make_toctree_node(
    state: docutils.parsers.rst.states.RSTState,
    toc_entries: List[Tuple[str, str]],
    options: dict,
    source_path: str,
    source_line: int = 0,
) -> List[docutils.nodes.Node]:
    if not toc_entries:
        return []

    # The Sphinx `toctree` directive parser cannot handle page names that
    # include angle brackets.  Therefore, we use the directive to create an
    # empty toctree node and then add the entries directly.
    toctree_nodes = parse_rst(
        state=state,
        text=format_directive("toctree", options=options),
        source_path=source_path,
        source_line=source_line,
    )
    toctree: Optional[sphinx.addnodes.toctree] = None
    for node in toctree_nodes[-1].findall(condition=sphinx.addnodes.toctree):
        toctree = node
        break
    if toctree is None:
        raise ValueError("No toctree node found")
    toctree["entries"].extend(toc_entries)
    toctree["includefiles"].extend([path for _, path in toc_entries])
    return toctree_nodes


def remove_css_file(app: sphinx.application.Sphinx, filename: str):
    """Removes a CSS file added by another extension."""
    app_css_files = app.registry.css_files
    app_indices = [i for i, x in enumerate(app_css_files) if x[0] == filename]
    for i in reversed(app_indices):
        del app_css_files[i]

    if hasattr(app, "builder") and hasattr(app.builder, "add_css_file"):
        builder_css_files = app.builder.css_files  # type: ignore[attr-defined]
        builder_indices = [
            i for i, x in enumerate(builder_css_files) if x.filename == filename
        ]
        for i in reversed(builder_indices):
            del builder_css_files[i]


@contextlib.contextmanager
def save_default_role(env: sphinx.environment.BuildEnvironment):
    orig_role_fn = docutils.parsers.rst.roles._roles.get("")  # type: ignore[attr-defined]
    orig_role_name = env.temp_data["default_role"]

    try:
        yield
    finally:
        if orig_role_fn is not None:
            docutils.parsers.rst.roles._roles[""] = orig_role_fn  # type: ignore[attr-defined]
        else:
            docutils.parsers.rst.roles._roles.pop("", None)  # type: ignore[attr-defined]
        env.temp_data["default_role"] = orig_role_name



"""
A special theme-specific extension to support "content tabs" from mkdocs-material.
"""
from typing import List
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective
from sphinx.application import Sphinx
from sphinx.util.logging import getLogger
from sphinx.writers.html import HTMLTranslator

LOGGER = getLogger(__name__)


def is_md_tab_type(node: nodes.Node, name: str):
    """Check if a node is a certain tabbed component."""
    if not isinstance(node, nodes.Element):
        return False
    return node.get("type") == name


class content_tab_set(nodes.container):
    pass


class MaterialTabSetDirective(SphinxDirective):
    """A container for a set of tab items."""

    has_content = True
    option_spec = {
        "name": directives.unchanged,
        "class": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()
        tab_set = content_tab_set(
            "",
            is_div=True,
            type="md-tab-set",
            classes=["tabbed-set", "tabbed-alternate"] + self.options.get("class", []),
        )
        self.set_source_info(tab_set)
        if self.options.get("name", ""):
            self.add_name(tab_set)
        self.state.nested_parse(self.content, self.content_offset, tab_set)
        for item in tab_set.children:
            if not is_md_tab_type(item, "md-tab-item"):
                LOGGER.warning(
                    "All children of a 'md-tab-set' should be 'md-tab-item'",
                    location=item,
                )
                break
        return [tab_set]


class MaterialTabItemDirective(SphinxDirective):
    """A single tab item in a tab set.
    Note: This directive generates a single container,
    for the label and content::
        <container design_component="tab-item" has_title=True>
            <rubric>
                ...title nodes
            <container design_component="tab-content">
                ...content nodes
    This allows for a default rendering in non-HTML outputs.
    The ``MaterialTabSetHtmlTransform`` then transforms this container
    into the HTML specific structure.
    """

    required_arguments = 1  # the tab label is the first argument
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        "class": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()
        if not is_md_tab_type(self.state_machine.node, "md-tab-set"):
            LOGGER.warning(
                "The parent of a 'md-tab-item' should be a 'md-tab-set'",
                location=(self.env.docname, self.lineno),
            )
        tab_item = nodes.container(
            "", is_div=True, type="md-tab-item", classes=["tabbed-block"]
        )

        # add tab label
        textnodes, _ = self.state.inline_text(self.arguments[0], self.lineno)
        tab_label = nodes.rubric("", "", *textnodes, classes=["tabbed-label"])
        self.add_name(tab_label)
        tab_item += tab_label

        # add tab content
        tab_content = nodes.container(
            "",
            is_div=True,
            type="",
            classes=["tabbed-block"] + self.options.get("class", []),
        )
        self.state.nested_parse(self.content, self.content_offset, tab_content)
        tab_item += tab_content

        return [tab_item]


class content_tab_label(nodes.TextElement, nodes.General):
    pass


def visit_tab_label(self, node):
    attributes = {"for": node["input_id"]}
    self.body.append(self.starttag(node, "label", **attributes))


def depart_tab_label(self, node):
    self.body.append("</label>")


def visit_tab_set(self: HTMLTranslator, node: content_tab_set):
    # increment tab set counter
    tab_set_count: int = getattr(self, "tab_set_count", 0) + 1
    setattr(self, "tab_set_count", tab_set_count)

    # configure tab set's div attributes
    tab_set_identity = f"__tabbed_{tab_set_count}"
    attributes = {"data-tabs": f"{tab_set_count}:{len(node.children)}"}
    self.body.append(self.starttag(node, "div", **attributes))

    # walkabout the children
    tab_label_div = nodes.container("", is_div=True, classes=["tabbed-labels"])
    tab_content_div = nodes.container("", is_div=True, classes=["tabbed-content"])
    for tab_count, tab_item in enumerate(node.children):
        assert isinstance(tab_item, nodes.Element)
        try:
            tab_label, tab_block = tab_item.children
        except ValueError as exc:
            raise ValueError(f"md-tab-item has no children:\n{repr(tab_item)}") from exc
        tab_item_identity = tab_set_identity + f"_{tab_count + 1}"

        assert isinstance(tab_label, nodes.Element)

        # create: <input checked="checked" id="id" type="radio">
        self.body.append(
            "<input "
            + ("checked " if not tab_count else "")
            + f'type="radio" id="{self.attval(tab_item_identity)}"'
            + f' name="{self.attval(tab_set_identity)}">'
        )

        # create: <label for="id">...</label>
        label_node = content_tab_label(
            "",
            "",
            *tab_label.children,
            input_id=tab_item_identity,
            classes=tab_label["classes"],
        )
        label_node.source, label_node.line = tab_item.source, tab_item.line
        tab_label_div += label_node

        # add content
        tab_content_div += tab_block

    tab_label_div.walkabout(self)
    tab_content_div.walkabout(self)
    self.body.append("</div>")
    raise nodes.SkipNode()


def depart_tab_set(self: HTMLTranslator, node: content_tab_set):
    pass


def setup(app: Sphinx):
    app.add_directive("md-tab-set", MaterialTabSetDirective)
    app.add_directive("md-tab-item", MaterialTabItemDirective)
    app.add_node(content_tab_label, html=(visit_tab_label, depart_tab_label))
    app.add_node(content_tab_set, html=(visit_tab_set, depart_tab_set))
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""This module adds the md-icon role. This is specific to this this theme because,
primarily, the icons bundled with this theme are supported (plus any svg files in the
user project's static directories)."""
from pathlib import Path
from typing import Tuple, List
from docutils import nodes
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.writers.html5 import HTML5Translator
from sphinx.util.docutils import SphinxRole


def load_svg_into_builder_env(builder: Builder, icon_name: str) -> str:
    """Reads the SVG data into a builder env variable for CSS generation.

    Returns: The name of CSS class associated with the SVG data."""
    css_icon_name = icon_name.replace("/", "_").replace("\\", "_")
    custom_icons = getattr(builder.env, "sphinx_immaterial_custom_icons", {})
    icon_name += ".svg"
    if css_icon_name not in custom_icons:
        static_paths: List[str] = getattr(builder.config, "sphinx_immaterial_icon_path")
        for path in static_paths:
            svg = Path(builder.srcdir) / path / icon_name
            if svg.exists():
                break
        else:
            svg = Path(__file__).parent / ".icons" / icon_name
            if not svg.exists():
                raise FileNotFoundError(
                    f"{icon_name} not found in sphinx_immaterial_icon_path and"
                    " not bundled with the theme"
                )
        custom_icons[css_icon_name] = svg.read_text(encoding="utf-8")
        setattr(builder.env, "sphinx_immaterial_custom_icons", custom_icons)
    return css_icon_name


class si_icon(nodes.container):
    pass


def visit_si_icon(self: HTML5Translator, node: si_icon):
    """Create a span element with associated CSS classes."""
    self.body.append(f'<span class="{" ".join(node["classes"])}"></span>')
    raise nodes.SkipNode()


class IconsRole(SphinxRole):
    """A interpreted text role to use icons in lines of text."""

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        path, classes = (self.text, "")
        if ";" in self.text:
            path, classes = self.text.split(";")[:2]
        class_list = [
            "md-icon",
            "si-icon-inline",
            load_svg_into_builder_env(
                builder=self.env.app.builder,
                icon_name=path,
            ),
        ]
        class_list.extend([cls for cls in classes.split(",") if cls])
        div = si_icon(path, classes=class_list)
        return [div], []


def setup(app: Sphinx):

    app.add_role("si-icon", IconsRole())
    app.add_node(si_icon, html=(visit_si_icon, None))

    app.add_config_value(
        name="sphinx_immaterial_icon_path",
        default=[],
        rebuild="env",
        types=[List[str]],
    )

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from typing import Optional

import docutils.nodes
import sphinx.domains.python
import sphinx.environment
import sphinx.addnodes


def _monkey_patch_python_parse_arglist():
    """Ensures default values in signatures are styled as code."""

    orig_parse_arglist = sphinx.domains.python._parse_arglist

    def parse_arglist(
        arglist: str, env: Optional[sphinx.environment.BuildEnvironment] = None
    ) -> sphinx.addnodes.desc_parameterlist:
        result = orig_parse_arglist(arglist, env)
        for node in result.findall(condition=docutils.nodes.inline):
            if "default_value" not in node["classes"]:
                continue
            node.replace_self(
                docutils.nodes.literal(
                    text=node.astext(),
                    classes=["code", "python", "default_value"],
                    language="python",
                )
            )
        return result

    sphinx.domains.python._parse_arglist = parse_arglist


_monkey_patch_python_parse_arglist()



import json
from typing import Tuple, cast, List

import docutils.nodes
import sphinx.ext.autodoc
import sphinx.addnodes
import sphinx.domains.python
from sphinx.domains.python import PyObject
from sphinx.domains.python import PythonDomain
import sphinx.util.logging

logger = sphinx.util.logging.getLogger(__name__)


def _monkey_patch_python_domain_to_support_object_ids():
    for object_class in sphinx.domains.python.PythonDomain.directives.values():
        object_class.option_spec["object-ids"] = json.loads
        object_class.option_spec["nonodeid"] = docutils.parsers.rst.directives.flag

    passthrough_options = ("object-ids", "nonodeid")

    orig_add_directive_header = sphinx.ext.autodoc.Documenter.add_directive_header

    def add_directive_header(self: sphinx.ext.autodoc.Documenter, sig: str) -> None:
        orig_add_directive_header(self, sig)
        for option_name in passthrough_options:
            if option_name not in self.options:
                continue
            value = self.options[option_name]
            self.add_line(f"   :{option_name}: {value}", self.get_sourcename())

    sphinx.ext.autodoc.Documenter.add_directive_header = add_directive_header

    orig_handle_signature = sphinx.domains.python.PyObject.handle_signature

    def handle_signature(
        self: sphinx.domains.python.PyObject,
        sig: str,
        signode: sphinx.addnodes.desc_signature,
    ) -> Tuple[str, str]:
        fullname, prefix = orig_handle_signature(self, sig, signode)
        object_ids = self.options.get("object-ids")
        if object_ids is not None:
            signature_index = getattr(self, "_signature_index", 0)
            setattr(self, "_signature_index", signature_index + 1)
            modname = signode["module"]
            if modname:
                modname += "."
            else:
                modname = ""
            if signature_index >= len(object_ids):
                logger.warning(
                    "Not enough object-ids %r specified for %r",
                    object_ids,
                    modname + signode["fullname"],
                    location=self.get_source_info(),
                )
            else:
                object_id = object_ids[signature_index]
                if object_id.startswith(modname):
                    fullname = object_id[len(modname) :]
                    signode["fullname"] = fullname
                else:
                    logger.warning(
                        "object-id %r for %r does not start with module name %r",
                        object_id,
                        signode["fullname"],
                        modname,
                        location=self.get_source_info(),
                    )
        return fullname, prefix

    sphinx.domains.python.PyObject.handle_signature = handle_signature

    orig_after_content = PyObject.after_content

    def after_content(self: PyObject) -> None:
        orig_after_content(self)
        obj_desc = cast(
            sphinx.addnodes.desc_content, getattr(self, "contentnode")
        ).parent
        signodes = obj_desc.children[:-1]

        py = cast(PythonDomain, self.env.get_domain("py"))

        def strip_object_entry_node_id(existing_node_id: str, object_id: str):
            obj = py.objects.get(object_id)
            if (
                obj is None
                or obj.node_id != existing_node_id
                or obj.docname != self.env.docname
            ):
                return
            py.objects[object_id] = obj._replace(node_id="")

        nonodeid = "nonodeid" in self.options
        canonical_name = self.options.get("canonical")
        noindexentry = "noindexentry" in self.options
        noindex = "noindex" in self.options

        for signode in cast(List[docutils.nodes.Element], signodes):
            modname = signode["module"]
            fullname = signode["fullname"]
            symbol = (modname + "." if modname else "") + fullname
            if nonodeid and signode["ids"]:
                orig_node_id = signode["ids"][0]
                signode["ids"] = []
                strip_object_entry_node_id(orig_node_id, symbol)
                if canonical_name:
                    strip_object_entry_node_id(orig_node_id, canonical_name)

                if noindexentry:
                    entries = cast(docutils.nodes.Element, self.indexnode)["entries"]
                    new_entries = []
                    for entry in entries:
                        new_entry = list(entry)
                        if new_entry[2] == orig_node_id:
                            new_entry[2] = ""
                        new_entries.append(tuple(new_entry))
                    cast(docutils.nodes.Element, self.indexnode)[
                        "entries"
                    ] = new_entries

    PyObject.after_content = after_content


_monkey_patch_python_domain_to_support_object_ids()



"""Modifies the sphinx search index.

- object synopses are added

- instead of the list of docnames, there is a list of URLs.  That way we don't
  need to duplicate in JavaScript the logic of determining a URL from a page
  name.

- the unused list of filenames is removed, since it just bloated the index.
"""

import io
from typing import cast, Any, Dict, IO, List, Tuple, Union, Optional

import sphinx
import sphinx.search
import sphinx.application
import sphinx.builders.html


def _get_all_synopses(
    env: sphinx.environment.BuildEnvironment,
) -> Dict[Tuple[Optional[str], str], str]:
    synopses: Dict[Tuple[Optional[str], str], str] = {}
    for domain_name, domain in env.domains.items():
        get_object_synopses = getattr(domain, "get_object_synopses", None)
        if get_object_synopses is not None:
            for key, synopsis in get_object_synopses():
                synopses.setdefault(key, synopsis or "")
        for (
            name,
            dispname,
            objtype,
            docname,
            anchor,
            priority,
        ) in domain.get_objects():
            synopses.setdefault((docname, anchor), "")
    return synopses


class IndexBuilder(sphinx.search.IndexBuilder):
    def get_objects(  # type: ignore[override]
        self, fn2index: Dict[str, int]
    ) -> Dict[
        str,
        Union[
            # From sphinx 4.3 onwards the children dict is now a list
            Dict[str, Tuple[int, int, int, Union[int, str]]],
            List[Tuple[int, int, int, Union[int, str], str]],
        ],
    ]:
        rv = super().get_objects(fn2index)
        onames = self._objnames
        docindex_to_docname = {i: docname for docname, i in fn2index.items()}
        synopses = _get_all_synopses(self.env)

        for prefix, prefix_value in rv.items():
            if prefix:
                name_prefix = prefix + "."
            else:
                name_prefix = ""
            if sphinx.version_info >= (4, 3):
                # From sphinx 4.3 onwards the children dict is now a list
                children = prefix_value
            else:
                children = [
                    (*values, name)  # type: ignore[misc]
                    for name, values in cast(dict, prefix_value).items()
                ]
            for i, (docindex, typeindex, prio, shortanchor, name) in enumerate(
                children
            ):
                objtype_entry = onames[typeindex]
                domain_name = objtype_entry[0]
                objtype = objtype_entry[1]
                domain = self.env.domains[domain_name]

                docname = docindex_to_docname.get(docindex)

                # Workaround for problems with the `shortanchor` encoding
                # https://github.com/sphinx-doc/sphinx/issues/10380

                # First decode `shortanchor` into the actual anchor based on the
                # full set of `(docname, anchor)` pairs.

                full_name = name_prefix + name

                anchor = shortanchor
                synopsis = synopses.get((docname, anchor))
                if shortanchor == "":
                    # The anchor could either be the actual empty string, or `full_name`.
                    if synopsis is None:
                        # Anchor is probably `full_name`
                        anchor = full_name
                        synopsis = synopses.get((docname, anchor))
                elif shortanchor == "-":
                    # The anchor could either by "-" or `objtype
                    if synopsis is None:
                        anchor = objtype + "-" + full_name
                        synopsis = synopses.get((docname, anchor))

                # Encode `anchor` into a lossless `shortanchor`.
                new_shortanchor: Union[int, str]
                if anchor == full_name:
                    new_shortanchor = 0
                elif anchor == objtype + "-" + full_name:
                    new_shortanchor = 1
                else:
                    new_shortanchor = anchor

                synopsis = synopsis or ""

                if sphinx.version_info >= (4, 3):
                    prefix_value[i] = (  # type: ignore
                        docindex,  # type: ignore
                        typeindex,
                        prio,
                        new_shortanchor,
                        name,
                        synopsis,
                    )
                else:
                    prefix_value[name] = (  # type: ignore
                        docindex,
                        typeindex,
                        prio,
                        new_shortanchor,
                        synopsis,
                    )
        return cast(Any, rv)

    def freeze(self):
        result = super().freeze()

        # filenames are unused
        result.pop("filenames")

        docnames = result.pop("docnames")

        builder = self.env.app.builder
        result.update(docurls=[builder.get_target_uri(docname) for docname in docnames])
        return result

    def load(
        self, stream: IO, format: Any  # pylint: disable=redefined-builtin
    ) -> None:
        if isinstance(format, str):
            format = self.formats[format]
        frozen = format.load(stream)
        # Reconstruct `docnames` and `filenames` from `docurls` since they are
        # expected by `IndexBuilder`.
        builder = self.env.app.builder
        assert isinstance(builder, sphinx.builders.html.StandaloneHTMLBuilder)
        url_to_docname = {
            builder.get_target_uri(docname): docname
            for docname in self.env.all_docs.keys()
        }
        docurls = frozen["docurls"]
        # Note: For documents that have been removed (and are therefore missing
        # from `url_to_docname`), `docnames` and `filenames` will just contain
        # the URL rather than the docname/filename.  However, since these will
        # be pruned anyway it does not hurt.
        docnames = []
        filenames = []
        for url in docurls:
            docname = url_to_docname.get(url, None)
            if docname is None:
                docnames.append(url)
                filenames.append(url)
            else:
                docnames.append(docname)
                filenames.append(self.env.doc2path(docname, False))
        frozen["docnames"] = docnames
        frozen["filenames"] = filenames
        new_data = format.dumps(frozen)
        if isinstance(new_data, str):
            stream = io.StringIO(new_data)
        else:
            stream = io.BytesIO(new_data)
        super().load(stream, format)


def _monkey_patch_index_builder():
    sphinx.search.IndexBuilder = IndexBuilder


def setup(app: sphinx.application.Sphinx):
    _monkey_patch_index_builder()
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Extension that adds `default-literal-role` directive."""

import contextlib
from typing import List

import docutils.nodes
from docutils.parsers.rst.states import Inliner
import sphinx.application
import sphinx.util.docutils
from sphinx.util.docutils import CustomReSTDispatcher

_LITERAL_KEY = "``"


def set_default_literal_role(role: str):
    def literal(self: Inliner, match, lineno: int):
        def nodeclass(rawsource: str, text: str):
            return self.interpreted(rawsource, text, role, lineno)

        nodeclass.__name__ = "literal"

        before, inlines, remaining, sysmessages, endstring = self.inline_obj(
            match, lineno, self.patterns.literal, nodeclass  # type: ignore[arg-type]
        )

        if len(inlines) == 1 and isinstance(inlines[0], tuple):
            inlines, extra_messages = inlines[0]
            sysmessages = sysmessages + extra_messages

        return before, inlines, remaining, sysmessages

    Inliner.dispatch[_LITERAL_KEY] = literal  # type: ignore[assignment]


@contextlib.contextmanager
def save_default_literal_role():
    orig_literal = Inliner.dispatch[_LITERAL_KEY]

    try:
        yield
    finally:
        Inliner.dispatch[_LITERAL_KEY] = orig_literal


@contextlib.contextmanager
def overridden_default_literal_role(role: str):
    """Overrides the role generated by double backticks literal syntax.

    Normally, double backticks:

        ``abc``

    is equivalent to the literal role:

        :literal:`abc`

    This context manager globally overrides the role to instead be the specified
    `role`.
    """

    with save_default_literal_role():
        set_default_literal_role(role)
        yield


class DefaultLiteralRoleDirective(sphinx.util.docutils.SphinxDirective):

    has_content = False
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = False

    def run(self) -> List[docutils.nodes.Node]:
        if not self.arguments:
            Inliner.dispatch[_LITERAL_KEY] = Inliner.literal  # type: ignore[assignment]
        else:
            set_default_literal_role(self.arguments[0])
        return []


def _monkey_patch_custom_rest_dispatcher():
    """Monkey patches `CustomReSTDispatcher` to also retore the default literal
    role.

    Note:

    The `sphinx_domains` context manager that wraps all document reading
    inherits from CustomReSTDispatcher.  Therefore, the modified
    `CustomReSTDispatcher` logic runs before and after each document is read.
    """

    orig_enable = CustomReSTDispatcher.enable

    def enable(self):
        orig_enable(self)
        setattr(self, "_literal_func", Inliner.dispatch[_LITERAL_KEY])

    CustomReSTDispatcher.enable = enable

    orig_disable = CustomReSTDispatcher.disable

    def disable(self):
        Inliner.dispatch[_LITERAL_KEY] = getattr(self, "_literal_func")
        orig_disable(self)

    CustomReSTDispatcher.disable = disable


_monkey_patch_custom_rest_dispatcher()


def setup(app: sphinx.application.Sphinx):
    app.add_directive("default-literal-role", DefaultLiteralRoleDirective)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from ._bravo import Bravo


class Alpha(Bravo):
    """
    Class Alpha docstring.
    """

    def my_method_Alpha(self) -> int:
        """
        This is my_method_Alpha.

        This is paragraph two.
        """
        return 1

    @property
    def my_property_Alpha(self) -> str:
        """
        This is my_property_Alpha.

        This is paragraph two.
        """
        return "alpha"



import docutils.nodes
import sphinx.application
import sphinx.addnodes

from . import object_description_options
from .. import html_translator_mixin


def _wrap_signature(node: sphinx.addnodes.desc_signature, limit: int):
    """Wraps long function signatures.

    Adds the `sig-wrap` class which causes each parameter to be displayed on a
    separate line.
    """
    node_text = node.astext()
    if len(node_text) > limit:
        node["classes"].append("sig-wrap")


def _wrap_signatures(
    app: sphinx.application.Sphinx,
    domain: str,
    objtype: str,
    content: docutils.nodes.Element,
) -> None:
    env = app.env
    assert env is not None
    options = object_description_options.get_object_description_options(
        env, domain, objtype
    )
    if (
        not options["wrap_signatures_with_css"]
        or options.get("clang_format_style") is not None
    ):
        return
    signatures = content.parent[:-1]
    for signature in signatures:
        assert isinstance(signature, sphinx.addnodes.desc_signature)
        _wrap_signature(signature, options["wrap_signatures_column_limit"])


@html_translator_mixin.override
def visit_desc_parameter(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.desc_parameter,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.desc_parameter],
) -> None:
    self.body.append('<span class="sig-param-decl">')
    super_func(self, node)


@html_translator_mixin.override
def depart_desc_parameter(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.desc_parameter,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.desc_parameter],
) -> None:
    super_func(self, node)
    self.body.append("</span>")


def setup(app: sphinx.application.Sphinx):
    app.setup_extension("sphinx_immaterial.apidoc.object_description_options")
    app.connect("object-description-transform", _wrap_signatures, priority=1000)

    object_description_options.add_object_description_option(
        app, "wrap_signatures_with_css", type_constraint=bool, default=True
    )
    object_description_options.add_object_description_option(
        app, "wrap_signatures_column_limit", type_constraint=int, default=68
    )

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Defines HTMLTranslatorMixin for overriding HTML translation.

Other extensions included with this theme add methods to the mixin.
"""

from typing import TYPE_CHECKING, TypeVar, Type, Callable, List

import docutils.nodes

import sphinx.writers.html5

if TYPE_CHECKING:
    HTMLTranslatorMixinBase = sphinx.writers.html5.HTML5Translator
else:
    HTMLTranslatorMixinBase = object


class HTMLTranslatorMixin(HTMLTranslatorMixinBase):  # pylint: disable=abstract-method
    pass


InitCallback = Callable[[HTMLTranslatorMixin], None]

_init_callbacks: List[InitCallback] = []


Element = TypeVar("Element", bound=docutils.nodes.Element)

BaseVisitCallback = Callable[
    [HTMLTranslatorMixin, Element],
    None,
]

VisitCallback = Callable[
    [
        HTMLTranslatorMixin,
        Element,
        BaseVisitCallback[Element],
    ],
    None,
]


def _override_visit_or_depart(method: str, callback: VisitCallback) -> None:
    prev_func = getattr(HTMLTranslatorMixin, method, None)

    def super_func(
        self: sphinx.writers.html5.HTML5Translator, node: docutils.nodes.Element
    ):
        if prev_func is not None:
            prev_func(self, node)
            return
        super_func = getattr(super(HTMLTranslatorMixin, self), method, None)
        if super_func is not None:
            super_func(node)

    def handler(self: HTMLTranslatorMixin, node: docutils.nodes.Element) -> None:
        callback(self, node, super_func)

    setattr(HTMLTranslatorMixin, method, handler)


def override(callback: VisitCallback):
    _override_visit_or_depart(callback.__name__, callback)


def init(callback: Callable[[HTMLTranslatorMixin], None]):
    _init_callbacks.append(callback)


def get_html_translator(
    base_translator: Type[sphinx.writers.html5.HTML5Translator],
) -> Type[sphinx.writers.html5.HTML5Translator]:
    class CustomHTMLTranslator(
        HTMLTranslatorMixin, base_translator  # type: ignore
    ):  # pylint: disable=abstract-method
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            # Ensure all tables are marked as data tables.  The material CSS only
            # applies to tables with this class, in order to leave tables used for
            # layout purposes alone.
            self.settings.table_style = ",".join(
                self.settings.table_style.split(",") + ["data"]
            )

            for callback in _init_callbacks:
                callback(self)

    return CustomHTMLTranslator



#! python
# type: ignore
from setuptools import setup
from pybind11.setup_helpers import Pybind11Extension, build_ext

ext_modules = [
    Pybind11Extension(
        "sphinx_immaterial_pybind11_issue_134",
        ["sphinx_immaterial_pybind11_issue_134.cpp"],
    ),
]

setup(
    name="sphinx-immaterial-pybind11-issue-134",
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
)



"""Demo file for Python API documentation generation.

This defines some entities that are aliased into the `tensorstore_demo` module.
"""

import typing


class Foo:
    """
    This is a class defined in the ``tensorstore_demo._tensorstore`` module
    but should appear to be defined in the ``tensorstore_demo`` module.

    Group:
      Some other group
    """

    def bar(self, x: int) -> int:
        """Returns :py:param:`x`.

        :param x: Parameter to :py:obj:`.bar`.
        """
        pass


def bar(x: Foo) -> Foo:
    """Returns :py:param:`x`.

    :param x: Parameter to :py:obj:`.bar`.

    Group:
      Some other group.
    """
    pass


class FooSubclass(Foo):
    """This is a subclass of :py:obj:`.Foo`.

    Group:
      Some other group.
    """

    def baz(self, x: str) -> str:
        """Does the baz operation."""
        return x

    @typing.overload
    def overloaded_method(self, x: int) -> int:
        pass

    @typing.overload
    def overloaded_method(self, x: str) -> str:
        pass

    def overloaded_method(self, x):
        """Returns its argument."""
        return x



"""Extends support for requires clauses in Sphinx C++ domain.

https://github.com/sphinx-doc/sphinx/pull/10286
"""

import re
from typing import Optional, Union

import sphinx.domains.cpp
from sphinx.domains.cpp import DefinitionParser, ASTDeclaration


def _monkey_patch_cpp_domain_support_requires_clause():
    orig_parse_declaration = DefinitionParser.parse_declaration

    orig_parse_type = DefinitionParser._parse_type

    def _parse_type(
        self, named: Union[bool, str], outer: Optional[str] = None
    ) -> sphinx.domains.cpp.ASTType:
        orig_assert_end = self.assert_end

        def assert_end_or_requires(allowSemicolon=False):
            # Allow trailing requires on constructors
            self.skip_ws()
            if re.compile(r"requires\b").match(self.definition, self.pos):
                return
            orig_assert_end(allowSemicolon=allowSemicolon)

        if outer == "function":
            try:
                self.assert_end = assert_end_or_requires
                return orig_parse_type(self, named, outer)
            finally:
                self.assert_end = orig_assert_end
        return orig_parse_type(self, named, outer)

    DefinitionParser._parse_type = _parse_type

    def parse_declaration(
        self: DefinitionParser, objectType: str, directiveType: str
    ) -> ASTDeclaration:
        if objectType not in ("type", "member", "class"):
            ast = orig_parse_declaration(self, objectType, directiveType)
            if (
                objectType == "function"
                and ast.templatePrefix is None
                and ast.trailingRequiresClause is None
            ):
                ast.trailingRequiresClause = self._parse_requires_clause()
                ast.semicolon = ast.semicolon or self.skip_string(";")
            return ast

        try:
            orig_parse_template_declaration_prefix = (
                self._parse_template_declaration_prefix
            )

            requires_clause = None

            def parse_template_declaration_prefix(*args, **kwargs):
                self._parse_template_declaration_prefix = (
                    orig_parse_template_declaration_prefix
                )
                result = orig_parse_template_declaration_prefix(*args, **kwargs)
                if result is None:
                    return None
                nonlocal requires_clause
                requires_clause = self._parse_requires_clause()
                return result

            self._parse_template_declaration_prefix = parse_template_declaration_prefix  # type: ignore[assignment]

            result = orig_parse_declaration(self, objectType, directiveType)
            result.requiresClause = requires_clause
            return result
        finally:
            self._parse_template_declaration_prefix = (  # type: ignore[assignment]
                orig_parse_template_declaration_prefix
            )

    DefinitionParser.parse_declaration = parse_declaration


_monkey_patch_cpp_domain_support_requires_clause()



import docutils.nodes
import sphinx.addnodes
import sphinx.application
import sphinx.config
from sphinx.locale import _

from .. import html_translator_mixin

# Add "highlight" class in order for pygments syntax highlighting CSS rules to
# apply to signatures.
if "highlight" not in sphinx.addnodes.desc_signature.classes:
    sphinx.addnodes.desc_signature.classes.append("highlight")


@html_translator_mixin.override
def visit_desc(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.desc,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.desc],
) -> None:
    # Object description node

    # These are converted to `<dl>` elements with the domain and objtype
    # as classes.

    # Augment the list of classes with `objdesc` to make it easier to
    # style these without resorting to hacks.
    node["classes"].append("objdesc")
    super_func(self, node)


@html_translator_mixin.override
def visit_desc_type(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.desc_type,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.desc_type],
) -> None:
    self.body.append(self.starttag(node, tagname="span", suffix="", CLASS="desctype"))


@html_translator_mixin.override
def depart_desc_type(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.desc_type,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.desc_type],
) -> None:
    self.body.append("</span>")


@html_translator_mixin.override
def visit_caption(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: docutils.nodes.caption,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.caption],
) -> None:
    attributes = {"class": "caption-text"}
    if isinstance(node.parent, docutils.nodes.container) and node.parent.get(
        "literal_block"
    ):
        # add highlight class to caption's div container.
        # This is needed to trigger mkdocs-material CSS rule `.highlight .filename`
        self.body.append('<div class="code-block-caption highlight">')
        # append a CSS class to trigger mkdocs-material theme's caption CSS style
        attributes["class"] += " filename"
    else:
        super_func(self, node)
    self.add_fignumber(node.parent)
    self.body.append(self.starttag(node, "span", **attributes))


@html_translator_mixin.override
def depart_caption(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: docutils.nodes.caption,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.caption],
) -> None:
    if not isinstance(node.parent, docutils.nodes.container) and not node.parent.get(
        "literal_block"
    ):
        # only append ending tag if parent is not a literal-block.
        # Because all elements in the caption should be within a span element
        self.body.append("</span>")

    # append permalink if available
    if isinstance(node.parent, docutils.nodes.container) and node.parent.get(
        "literal_block"
    ):
        self.add_permalink_ref(node.parent, _("Permalink to this code"))
        self.body.append("</span>")  # done; add closing tag
    elif isinstance(node.parent, docutils.nodes.figure):
        self.add_permalink_ref(node.parent, _("Permalink to this image"))
    elif node.parent.get("toctree"):
        self.add_permalink_ref(node.parent.parent, _("Permalink to this toctree"))

    if isinstance(node.parent, docutils.nodes.container) and node.parent.get(
        "literal_block"
    ):
        self.body.append("</div>\n")
    elif not isinstance(node.parent, docutils.nodes.figure):
        # calling super_func() on a figure results in 2 permalinks in caption
        super_func(self, node)


# `desc_inline` nodes are generated by the `cpp:expr` role.
#
# Wrap it in a `<code>` element with the "highlight" class to ensure it
# displays properly as an inline code literal.
@html_translator_mixin.override
def visit_desc_inline(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.desc_inline,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.desc_inline],
) -> None:
    self.body.append(self.starttag(node, tagname="code", suffix="", CLASS="highlight"))


@html_translator_mixin.override
def depart_desc_inline(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.desc_inline,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.desc_inline],
) -> None:
    self.body.append("</code>")


@html_translator_mixin.override
def visit_productionlist(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: sphinx.addnodes.productionlist,
    super_func: html_translator_mixin.BaseVisitCallback[sphinx.addnodes.productionlist],
):
    # we only need to patch in a code element around the production list
    before_len = len(self.body)
    try:
        super_func(self, node)
    except docutils.nodes.SkipNode:
        pass
    # be sure that there wasn't an incompatible change to sphinx
    assert self.body[before_len].endswith("<pre>\n") and self.body[-1].startswith(
        "</pre>"
    )
    self.body[before_len] += "<code>"
    self.body[-1] = "</code>" + self.body[-1]
    raise docutils.nodes.SkipNode()


@html_translator_mixin.init
def init_translator(self: html_translator_mixin.HTMLTranslatorMixin) -> None:

    # Ensure pygments uses <code> elements, for compatibility with the
    # mkdocs-material CSS which expects that.
    self.highlighter.formatter_args.update(wrapcode=True)

    # Ensure classes like `s` (used for string literals in code
    # highlighting) aren't converted to `<s>` elements (strikethrough).
    # Sphinx already overrides this, but for some reason due to
    # `__init__` invocation order it gets overridden.
    self.supported_inline_tags = set()


def _config_inited(
    app: sphinx.application.Sphinx, config: sphinx.config.Config
) -> None:
    # make code-blocks' line numbers be a separate column of a 1-row table
    config["html_codeblock_linenos_style"] = "table"  # default is "inline"


def setup(app: sphinx.application.Sphinx):
    app.connect("config-inited", _config_inited)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import sphinx.application

from . import ast_fixes  # type: ignore[unused-import]
from . import cpp_resolve_c_xrefs  # type: ignore[unused-import]
from . import include_directives_in_signatures  # type: ignore[unused-import]
from . import last_resolved_symbol  # type: ignore[unused-import]
from . import macro_parameters  # type: ignore[unused-import]
from . import parameter_objects  # type: ignore[unused-import]
from . import signodes  # type: ignore[unused-import]
from . import strip_namespaces_from_signatures  # type: ignore[unused-import]
from . import symbol_ids  # type: ignore[unused-import]
from . import synopses  # type: ignore[unused-import]


def setup(app: sphinx.application.Sphinx):
    app.setup_extension(signodes.__name__)
    app.setup_extension(strip_namespaces_from_signatures.__name__)
    app.setup_extension(cpp_resolve_c_xrefs.__name__)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Test specific to https://github.com/jbms/sphinx-immaterial/issues/134
"""
from typing import List
import pytest
from sphinx.testing.util import SphinxTestApp
from sphinx_immaterial_pybind11_issue_134 import Example  # type: ignore


@pytest.mark.parametrize(
    "args,expected", [([], False), ([True], True)], ids=["without arg", "with arg"]
)
def test_issue_134_example_pkg(args: List[bool], expected: bool):
    cls = Example(*args)
    assert cls.is_set_by_init is expected


def test_autoclass_members(immaterial_make_app):
    """Issue 134 is related to using ``:members:`` option of
    the ``autoclass`` directive."""
    app: SphinxTestApp = immaterial_make_app(
        files={
            "index.rst": "\n".join(
                [
                    "Some API",
                    "========",
                    ".. autoclass:: sphinx_immaterial_pybind11_issue_134.Example",
                    "    :members: is_set_by_init",
                    "    :special-members: __init__",  # tests overloads in pybind11 (for good measure)
                ]
            )
        },
        extra_conf='extensions.append("sphinx.ext.autodoc")\n',
    )
    app.build()
    assert not app._warning.getvalue()
    print(app._status.getvalue())



"""Support for cross-referencing C and C++ symbols to cppreference.com."""

import importlib.resources
import os
import pathlib
import re
import typing
import urllib.parse

import xml.etree.ElementTree as ET
import sphinx.application

from . import external_cpp_references


def _join_links(a: str, b: typing.Optional[str]):
    if b is None or b == ".":
        return a
    return os.path.join(a, b)


def _join_names(a: str, b: str):
    if not a:
        return b
    return f"{a}::{b}"


def _get_last_name(s: str) -> str:
    return re.sub(".*::", "", s)


class CppReferenceXmlParser:
    def __init__(
        self,
        cpp_objects: typing.Dict[str, external_cpp_references.ObjectInfo],
        base_url: str,
    ):
        self.cpp_objects = cpp_objects
        self.base_url = base_url
        self.name_prefix = ""
        self.link_prefix = ""
        self.since = ""

    def add_file(self, content: bytes, since: str):
        content = content.replace(b"<>", b"&lt;&gt;")
        root = ET.fromstring(content)
        self.since = since
        for child in root:
            self.add_element(child)

    def add_url(self, name: str, typ: str, url: str, since: str):
        self.cpp_objects[name] = external_cpp_references.ObjectInfo(
            url=url,
            desc=f"{name} ({since.upper()} standard {typ})",
            object_type=typ,
        )

    def add(self, name: str, typ: str, link: str, since: str):
        self.add_url(
            name=name,
            typ=typ,
            url=urllib.parse.urljoin(self.base_url, link),
            since=since,
        )

    def add_simple(self, element: ET.Element, typ: str):
        name = element.get("name")
        assert name is not None
        since = element.get("since", self.since)
        self.add(
            name=_join_names(self.name_prefix, name),
            typ=typ,
            link=_join_links(self.link_prefix, element.get("link", name)),
            since=since,
        )

    def add_element(self, element: ET.Element):
        add_function = getattr(self, f"_add_from_{element.tag}", None)
        if not add_function:
            return
        add_function(element)

    def _add_from_typedef(self, element: ET.Element):
        alias = element.get("alias")
        name = element.get("name")
        assert name is not None
        if alias is not None:
            existing_obj = self.cpp_objects[alias]
            full_name = _join_names(self.name_prefix, name)
            self.add_url(
                name=full_name,
                url=existing_obj.url,
                since=element.get("since", self.since),
                typ="type alias",
            )
            return
        self.add_simple(element, "type alias")

    def _add_from_class(self, element: ET.Element):
        name = element.get("name")
        assert name is not None
        prev_name_prefix = self.name_prefix
        prev_link_prefix = self.link_prefix
        prev_since = self.since

        self.name_prefix = _join_names(self.name_prefix, name)
        self.link_prefix = _join_links(self.link_prefix, element.get("link", name))
        self.since = element.get("since", self.since)
        self.add(
            name=self.name_prefix, typ="class", link=self.link_prefix, since=self.since
        )
        for child in element:
            self.add_element(child)
        self.name_prefix = prev_name_prefix
        self.link_prefix = prev_link_prefix
        self.since = prev_since

    def _add_from_const(self, element: ET.Element):
        self.add_simple(element, "var")

    def _add_from_enum(self, element: ET.Element):
        self.add_simple(element, "enum")

    def _add_from_function(self, element: ET.Element):
        self.add_simple(element, "function")

    def _add_from_overload(self, element: ET.Element):
        # Skip hidden friends for now
        pass

    def _add_from_variable(self, element: ET.Element):
        self.add_simple(element, "var")

    def _add_from_constructor(self, element: ET.Element):
        name = _get_last_name(self.name_prefix)
        self.add(
            name=_join_names(self.name_prefix, name),
            typ="function",
            link=_join_links(self.link_prefix, element.get("link")),
            since=element.get("since", self.since),
        )

    def _add_from_destructor(self, element: ET.Element):
        name = _get_last_name(self.name_prefix)
        self.add(
            name=_join_names(self.name_prefix, "~" + name),
            typ="function",
            link=_join_links(self.link_prefix, element.get("link")),
            since=element.get("since", self.since),
        )


def load_mappings(app: sphinx.application.Sphinx):
    parser = CppReferenceXmlParser(
        external_cpp_references.get_mappings(app),
        base_url="https://en.cppreference.com/w/",
    )
    xml_files = app.config.cppreference_xml_files
    if xml_files is not None:
        # Use user-specified XML files.
        xml_content = [
            (standard, pathlib.Path(os.path.join(app.srcdir, filename)).read_bytes())
            for standard, filename in app.config.cppreference_xml_files
        ]
    else:
        # Use default XML files included with extension.
        xml_content = [
            (
                "C",
                importlib.resources.read_binary(
                    "sphinx_immaterial.apidoc.cpp.cppreference_data",
                    "index-functions-c.xml",
                ),
            ),
            (
                "C++",
                importlib.resources.read_binary(
                    "sphinx_immaterial.apidoc.cpp.cppreference_data",
                    "index-functions-cpp.xml",
                ),
            ),
        ]
    for standard, content in xml_content:
        parser.add_file(content, since=standard)


def setup(app: sphinx.application.Sphinx):
    app.setup_extension("sphinx_immaterial.apidoc.cpp.external_cpp_references")
    app.connect("builder-inited", load_mappings)
    app.add_config_value(
        name="cppreference_xml_files",
        default=None,
        types=(typing.Optional[typing.List[typing.Tuple[str, str]]],),
        rebuild="env",
    )
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Adds support for :noindex:, :symbol-ids:, and :node-id: options."""

import json
from typing import Optional, Union, Type, List

import docutils.parsers.rst.directives
import sphinx.addnodes
import sphinx.domains.c
import sphinx.domains.cpp

AST_ID_OVERRIDE_ATTR = "_sphinx_immaterial_id"
ANCHOR_ATTR = "sphinx_immaterial_anchor"


def _monkey_patch_override_ast_id(
    ast_declaration_class: Union[
        Type[sphinx.domains.c.ASTDeclaration], Type[sphinx.domains.cpp.ASTDeclaration]
    ]
):
    """Allows the Symbol id to be overridden."""
    orig_get_id = ast_declaration_class.get_id

    def get_id(
        self: Union[sphinx.domains.c.ASTDeclaration, sphinx.domains.cpp.ASTDeclaration],
        version: int,
        prefixed: bool = True,
    ):
        entity_id = getattr(self, AST_ID_OVERRIDE_ATTR, None)
        if entity_id is not None:
            return entity_id
        return orig_get_id(self, version, prefixed)  # type: ignore

    ast_declaration_class.get_id = get_id  # type: ignore


def get_symbol_anchor(
    symbol: Union[sphinx.domains.c.Symbol, sphinx.domains.cpp.Symbol]
) -> str:
    anchor = getattr(symbol.declaration, ANCHOR_ATTR, None)
    if anchor is None:
        assert symbol.declaration is not None
        anchor = symbol.declaration.get_newest_id()
    return anchor


def _monkey_patch_cpp_noindex_option(
    object_class: Union[
        Type[sphinx.domains.c.CObject], Type[sphinx.domains.cpp.CPPObject]
    ],
    ast_declaration_class: Union[
        Type[sphinx.domains.c.ASTDeclaration], Type[sphinx.domains.cpp.ASTDeclaration]
    ],
    env_parent_symbol_key: str,
    duplicate_symbol_error: Union[
        Type[sphinx.domains.c._DuplicateSymbolError],
        Type[sphinx.domains.cpp._DuplicateSymbolError],
    ],
):
    object_class.option_spec["noindex"] = docutils.parsers.rst.directives.flag
    object_class.option_spec["symbol-ids"] = json.loads
    object_class.option_spec["node-id"] = lambda s: s or ""
    orig_handle_signature = object_class.handle_signature

    NEW_SYMBOLS_ATTR = "_sphinx_immaterial_new_symbols"

    SIGNATURE_INDEX_ATTR = "_sphinx_immaterial_signature_index"

    def handle_signature(
        self: Union[sphinx.domains.c.CObject, sphinx.domains.cpp.CPPObject],
        sig: str,
        signode: sphinx.addnodes.desc_signature,
    ):
        signature_index = getattr(self, SIGNATURE_INDEX_ATTR, -1) + 1
        setattr(self, SIGNATURE_INDEX_ATTR, signature_index)

        parentSymbol = self.env.temp_data[env_parent_symbol_key]
        orig_add_declaration = parentSymbol.add_declaration

        # For the `noindex` option, keep track of symbols that are added so that
        # they can be removed.
        new_symbols = getattr(self, NEW_SYMBOLS_ATTR, None)
        if new_symbols is None:
            new_symbols = []
            setattr(self, NEW_SYMBOLS_ATTR, new_symbols)

        symbol_ids = self.options.get("symbol-ids")
        symbol_id: Optional[str] = None
        if symbol_ids and signature_index < len(symbol_ids):
            symbol_id = symbol_ids[signature_index]

        def add_declaration(declaration, docname: str, line: int):
            if symbol_id is not None:
                setattr(declaration, AST_ID_OVERRIDE_ATTR, symbol_id)
            try:
                symbol = orig_add_declaration(declaration, docname, line)
                assert new_symbols is not None
                new_symbols.append(symbol)
            except duplicate_symbol_error as e:
                # Ignore duplicate symbols
                symbol = e.symbol
                declaration.symbol = symbol
                # Temporarily clear siblingAbove/siblingBelow to avoid an assertion
                # failure in `orig_handle_signature`.  They will be restored by
                # `orig_handle_signature`.
                siblingBelow = symbol.siblingBelow
                siblingAbove = symbol.siblingAbove
                if siblingAbove is not None:
                    if siblingAbove is not symbol:
                        symbol.siblingAbove.siblingBelow = siblingBelow
                    symbol.siblingAbove = None
                if siblingBelow is not None:
                    if siblingBelow is not symbol:
                        symbol.siblingBelow.siblingAbove = siblingAbove
                    symbol.siblingBelow = None

                # Remove duplicate symbol that was just created
                for other_symbol in symbol.parent._children:
                    if other_symbol.declaration is declaration:
                        assert other_symbol.isRedeclaration
                        other_symbol.remove()
                        break
                else:
                    raise AssertionError(  # pylint: disable=raise-missing-from
                        "Duplicate symbol not found: %r" % (symbol.dump(2),)
                    )
            return symbol

        parentSymbol.add_declaration = add_declaration
        try:
            return orig_handle_signature(self, sig, signode)  # type: ignore[arg-type]
        finally:
            parentSymbol.add_declaration = orig_add_declaration

    object_class.handle_signature = handle_signature  # type: ignore[assignment]

    orig_add_target_and_index = object_class.add_target_and_index

    def add_target_and_index(
        self: Union[sphinx.domains.c.CObject, sphinx.domains.cpp.CPPObject],
        ast: Union[sphinx.domains.c.ASTDeclaration, sphinx.domains.cpp.ASTDeclaration],
        sig: str,
        signode: sphinx.addnodes.desc_signature,
    ) -> None:
        node_id = self.options.get("node-id")
        if node_id is not None:
            symbol = ast.symbol
            setattr(symbol.declaration, ANCHOR_ATTR, node_id)
            symbol.docname = self.env.docname
            if ast is self.names[0] and node_id:
                signode["ids"].append(node_id)
            return

        orig_add_target_and_index(self, ast, sig, signode)  # type: ignore[arg-type]

    object_class.add_target_and_index = add_target_and_index  # type: ignore[assignment]

    orig_run = object_class.run

    def run(
        self: Union[sphinx.domains.c.CObject, sphinx.domains.cpp.CPPObject]
    ) -> List[docutils.nodes.Node]:
        result = orig_run(self)  # type: ignore[arg-type]

        if "noindex" in self.options:
            new_symbols = getattr(self, NEW_SYMBOLS_ATTR, None)
            if new_symbols:
                for s in new_symbols:
                    s.remove()

        return result

    object_class.run = run  # type: ignore[assignment]


_monkey_patch_override_ast_id(sphinx.domains.c.ASTDeclaration)
_monkey_patch_override_ast_id(sphinx.domains.cpp.ASTDeclaration)
_monkey_patch_cpp_noindex_option(
    object_class=sphinx.domains.cpp.CPPObject,
    ast_declaration_class=sphinx.domains.cpp.ASTDeclaration,
    env_parent_symbol_key="cpp:parent_symbol",
    duplicate_symbol_error=sphinx.domains.cpp._DuplicateSymbolError,
)
_monkey_patch_cpp_noindex_option(
    object_class=sphinx.domains.c.CObject,
    ast_declaration_class=sphinx.domains.c.ASTDeclaration,
    env_parent_symbol_key="c:parent_symbol",
    duplicate_symbol_error=sphinx.domains.c._DuplicateSymbolError,
)



class Foo:
    @classmethod
    def my_classmethod(cls, arg: str) -> int:
        return 1

    @staticmethod
    def my_staticmethod(arg: str) -> int:
        return 1

    def my_method(self, arg: str) -> int:
        return 1



"""Adds a new directive for using upstream's code annotations feature.
See comments in visit_literal_block() override below for detailed explanation about
why this directive is needed."""
from typing import List, cast
from docutils import nodes
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective
from sphinx.writers.html import HTMLTranslator
from sphinx.util.logging import getLogger
from . import html_translator_mixin

LOGGER = getLogger(__name__)


class annotations_list(nodes.container):
    pass


class CodeAnnotations(SphinxDirective):
    """Take a list of annotations and move them into a previous node's parent."""

    has_content = True

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()

        div = annotations_list("", used=False)
        self.state.nested_parse(self.content, self.content_offset, div)
        self.set_source_info(div)
        if not isinstance(div.children[0], nodes.enumerated_list):
            LOGGER.error(
                "The code-annotations directive only accepts an enumerated list.",
                location=div,
            )
        return [div]


def visit_annotations_list(self: HTMLTranslator, node: annotations_list):
    # the content should have been moved into the literal block before calling this
    if not node["used"]:  # see visit_literal_block() override below
        node.children[0].walkabout(self)
    raise nodes.SkipNode()


@html_translator_mixin.override
def visit_literal_block(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: nodes.literal_block,
    super_func: html_translator_mixin.BaseVisitCallback[nodes.literal_block],
):
    #  we need to be sure that the annotated_list is the next node
    annotation_parent = cast(nodes.Element, node)
    if isinstance(node.parent.children[0], nodes.caption):
        # literal blocks with a caption have added parent
        annotation_parent = node.parent
    next_element = annotation_parent.next_node(
        nodes.Element, siblings=True, descend=False
    )
    if isinstance(next_element, annotations_list):
        try:
            super_func(self, node)
        except nodes.SkipNode:
            pass  # we're not done yet
        # the last line appended to self.body should now have the entire code-block:
        # <div class="highlight-xxxxx notranslate">
        #     <div class="highlight">
        #         <pre><span></span><code>...</code></pre>
        #     </div>
        #     <!-- we need to insert annotations' list here -->
        # </div>
        assert (  # to try detecting future changes in sphinx/pygments
            self.body[-1].endswith("</div>\n")
            and '<div class="highlight">' in self.body[-1]
        )
        # Add annotate class to the div.highlight tag (generated by pygments), and
        # remove the closing div tag (`[:-7]`) of the div.highlight's parent node.
        self.body[-1] = self.body[-1][:-7].replace(
            '<div class="highlight">', '<div class="annotate highlight">'
        )
        next_element.children[0].walkabout(self)  # add annotations' enumerated_list
        self.body.append("</div>\n")  # re-add the closing div for the parent
        next_element["used"] = True
        raise nodes.SkipNode()  # now we're done
    # else no annotations list found; proceed like normal
    super_func(self, node)


def setup(app: Sphinx):
    app.add_directive("code-annotations", CodeAnnotations)
    app.add_node(annotations_list, html=(visit_annotations_list, None))
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Adds all Sphinx "objects" to the table of contents."""

from typing import cast, Optional, Union, Any, List
import docutils.nodes
import sphinx.addnodes
import sphinx.application
import sphinx.directives
from sphinx.environment.collectors.toctree import TocTreeCollector
from sphinx.locale import _

from . import object_description_options
from .. import html_translator_mixin


def _monkey_patch_toc_tree_process_doc():
    """Enables support for also finding Sphinx domain objects."""

    # Apply the monkey patch
    orig_process_doc = TocTreeCollector.process_doc

    def _make_section_from_desc(
        app: sphinx.application.Sphinx,
        source: sphinx.addnodes.desc,
    ) -> Optional[docutils.nodes.section]:
        env = app.env
        assert env is not None
        options = object_description_options.get_object_description_options(
            env, source["domain"], source["objtype"]
        )
        if not options["include_in_toc"]:
            return None

        signature: sphinx.addnodes.desc_signature
        for child in source.children:
            if not isinstance(child, sphinx.addnodes.desc_signature):
                continue
            signature = child
            break
        else:
            # No signature found
            return None
        ids = signature["ids"]
        if not ids:
            # Not indexed.
            return None
        section = docutils.nodes.section()
        section["ids"] = ids

        # Extract title from signature
        title = signature.get("toc_title", None)
        if not title:
            title = ""
            for child in signature.findall():
                if isinstance(
                    child, (sphinx.addnodes.desc_name, sphinx.addnodes.desc_addname)
                ):
                    child = cast(
                        Union[sphinx.addnodes.desc_name, sphinx.addnodes.desc_addname],
                        child,
                    )
                    if "sig-name-nonprimary" in child["classes"]:
                        continue
                    title += child.astext()
        if not title:
            # No name found
            return None
        # Sphinx uses the first child of the section node as the title.
        titlenode = docutils.nodes.comment(title, title)
        section += titlenode
        return section

    def _make_section_from_field(
        source: docutils.nodes.field,
    ) -> Optional[docutils.nodes.section]:
        fieldname = cast(docutils.nodes.field_name, source[0])
        fieldbody = cast(docutils.nodes.field_body, source[1])
        ids = fieldname["ids"]
        if not ids:
            # Not indexed
            return None
        section = docutils.nodes.section()
        section["ids"] = ids
        title = fieldname.astext()
        # Sphinx uses the first child of the section node as the title.
        titlenode = docutils.nodes.comment(title, title)
        section += titlenode
        return section

    def _make_section_from_rubric(
        source: docutils.nodes.rubric,
    ) -> Optional[docutils.nodes.section]:
        rubric = cast(docutils.nodes.rubric, source)
        ids = rubric["ids"]
        if not ids:
            # Not indexed
            return None
        section = docutils.nodes.section()
        section["ids"] = ids
        title = rubric.astext()
        # Sphinx uses the first child of the section node as the title.
        titlenode = docutils.nodes.comment(title, title)
        section += titlenode
        return section

    def _make_section_from_term(
        source: docutils.nodes.term,
    ) -> Optional[docutils.nodes.section]:
        ids = source["ids"]
        if not ids:
            # Not indexed
            return None
        section = docutils.nodes.section()
        section["ids"] = ids
        title = source["toc_title"]
        titlenode = docutils.nodes.comment(title, title)
        section += titlenode
        return section

    def _patched_process_doc(
        self: TocTreeCollector,
        app: sphinx.application.Sphinx,
        doctree: docutils.nodes.document,
    ) -> None:
        new_document = doctree.copy()  # Shallow copy

        def _collect(
            source: docutils.nodes.Node, target: docutils.nodes.Element
        ) -> None:
            if not isinstance(source, docutils.nodes.Element):
                return
            children = iter(source.children)
            new_node: Any
            if isinstance(source, docutils.nodes.section):
                new_node = source.copy()
                # Also copy first child, which sphinx interprets as the title
                new_node += next(children).deepcopy()
                target += new_node
                target = new_node
            elif isinstance(source, sphinx.addnodes.only):
                # Retain "only" nodes since they affect toc generation.
                new_node = source.copy()
                target += new_node
                target = new_node
            elif isinstance(source, sphinx.addnodes.toctree):
                # Deep copy entire toctree
                new_node = source.deepcopy()
                target += new_node
                return
            elif isinstance(source, sphinx.addnodes.desc):
                # Object description.  Try to create synthetic section.
                new_node = _make_section_from_desc(app, source)
                if new_node is not None:
                    target += new_node
                    target = new_node
            elif isinstance(source, docutils.nodes.field):
                # Field.  Try to create synthetic section.
                new_node = _make_section_from_field(source)
                if new_node is not None:
                    target += new_node
                    target = new_node
            elif isinstance(source, docutils.nodes.rubric):
                # Rubric.  Try to create synthetic section.
                new_node = _make_section_from_rubric(source)
                if new_node is not None:
                    target += new_node
                # Rubrics cannot contain sub-sections
                return
            elif isinstance(source, docutils.nodes.term) and source.get("toc_title"):
                # Term with toc title.  Try to create synthetic section.
                new_node = _make_section_from_term(source)
                if new_node is not None:
                    target += new_node
                # Parameters cannot contain sub-sections
                return

            for child in children:
                _collect(child, target)

        _collect(doctree, new_document)
        return orig_process_doc(self, app, new_document)

    TocTreeCollector.process_doc = _patched_process_doc  # type: ignore


_monkey_patch_toc_tree_process_doc()


@html_translator_mixin.override
def depart_field_name(
    self,
    node: docutils.nodes.field_name,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.field_name],
) -> None:
    super_func(self, node)
    last_part = self.body[-1]
    end_tag_index = last_part.index("</dt>")
    self.body[-1] = last_part[:end_tag_index]
    self.add_permalink_ref(node, _("Permalink to this headline"))
    self.body.append(last_part[end_tag_index:])


def _monkey_patch_object_description_to_include_fields_in_toc():
    orig_run = sphinx.directives.ObjectDescription.run

    def run(self: sphinx.directives.ObjectDescription) -> List[docutils.nodes.Node]:
        nodes = orig_run(self)

        options = object_description_options.get_object_description_options(
            self.env, self.domain, self.objtype
        )
        if not options["include_fields_in_toc"]:
            return nodes

        obj_desc = nodes[-1]

        obj_id = None
        for sig in obj_desc[:-1]:
            ids = sig["ids"]
            if ids and ids[0]:
                obj_id = ids[0]
                break

        obj_content = obj_desc[-1]
        for child in obj_content:
            if not isinstance(child, docutils.nodes.field_list):
                continue
            for field in child:
                assert isinstance(field, docutils.nodes.field)
                field_name = cast(docutils.nodes.field_name, field[0])
                if field_name["ids"]:
                    continue
                field_id = docutils.nodes.make_id(field_name.astext())
                if obj_id:
                    field_id = f"{obj_id}-{field_id}"
                field_name["ids"].append(field_id)

        return nodes

    sphinx.directives.ObjectDescription.run = run


_monkey_patch_object_description_to_include_fields_in_toc()


@html_translator_mixin.override
def depart_rubric(
    self,
    node: docutils.nodes.rubric,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.rubric],
) -> None:
    self.add_permalink_ref(node, _("Permalink to this headline"))
    super_func(self, node)


def _monkey_patch_object_description_to_include_rubrics_in_toc():
    orig_run = sphinx.directives.ObjectDescription.run

    def run(self: sphinx.directives.ObjectDescription) -> List[docutils.nodes.Node]:
        nodes = orig_run(self)

        options = object_description_options.get_object_description_options(
            self.env, self.domain, self.objtype
        )
        if not options["include_rubrics_in_toc"]:
            return nodes

        obj_desc = nodes[-1]
        obj_content = obj_desc[-1]
        for child in obj_content:
            if not isinstance(child, docutils.nodes.rubric):
                continue
            rubric = cast(docutils.nodes.rubric, child)
            if rubric["ids"]:
                continue
            rubric_id = docutils.nodes.make_id(rubric.astext())
            rubric["ids"].append(rubric_id)

        return nodes

    sphinx.directives.ObjectDescription.run = run


_monkey_patch_object_description_to_include_rubrics_in_toc()


@html_translator_mixin.override
def depart_term(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: docutils.nodes.term,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.term],
) -> None:
    self.add_permalink_ref(node, _("Permalink to this definition"))
    super_func(self, node)


def setup(app: sphinx.application.Sphinx):
    app.setup_extension("sphinx_immaterial.apidoc.object_description_options")
    object_description_options.add_object_description_option(
        app, "include_in_toc", type_constraint=bool, default=True
    )
    object_description_options.add_object_description_option(
        app, "toc_icon_text", type_constraint=Optional[str], default=None
    )
    object_description_options.add_object_description_option(
        app, "toc_icon_class", type_constraint=Optional[str], default=None
    )
    object_description_options.add_object_description_option(
        app, "include_fields_in_toc", type_constraint=bool, default=True
    )
    object_description_options.add_object_description_option(
        app, "include_rubrics_in_toc", type_constraint=bool, default=False
    )

    # TocTreeCollector is registered before our extension is.  In order for the
    # monkey patching to take effect, we need to unregister it and re-register it.
    for read_listener in app.events.listeners["doctree-read"]:
        obj = getattr(read_listener.handler, "__self__", None)
        if obj is not None and isinstance(obj, TocTreeCollector):
            obj.disable(app)
            app.add_env_collector(TocTreeCollector)
            break
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }
